# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import inspect
import os
import random
from collections import ChainMap
from typing import Any, Dict, Iterable, List, Optional, Set, Union

import numpy as np
import tensorflow as tf
import torch
from tensorflow.python.distribute.input_lib import DistributedDataset
from torch.utils.data import DataLoader

import fastestimator as fe
from fastestimator.backend.to_shape import to_shape
from fastestimator.backend.to_tensor import to_tensor
from fastestimator.backend.to_type import to_type
from fastestimator.dataset.op_dataset import OpDataset
from fastestimator.network import BaseNetwork, TFNetwork, TorchNetwork
from fastestimator.pipeline import Pipeline
from fastestimator.schedule.schedule import Scheduler, get_current_items, get_signature_epochs
from fastestimator.summary.history import HistoryRecorder
from fastestimator.summary.system import Summary, System
from fastestimator.trace.io.best_model_saver import BestModelSaver
from fastestimator.trace.io.model_saver import ModelSaver
from fastestimator.trace.io.restore_wizard import RestoreWizard
from fastestimator.trace.io.traceability import Traceability
from fastestimator.trace.trace import EvalEssential, Logger, PerDSTrace, TestEssential, Trace, TrainEssential, \
    sort_traces
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import NonContext, Suppressor, draw, to_list, to_set


@traceable()
class Estimator:
    """One class to rule them all.

    Estimator is the highest level class within FastEstimator. It is the class which is invoked to actually train
    (estimator.fit) or test (estimator.test) models. It wraps `Pipeline`, `Network`, `Trace` objects together and
    defines the whole optimization process.

    Args:
        pipeline: An fe.Pipeline object that defines the data processing workflow.
        network: An fe.Network object that contains models and other training graph definitions.
        epochs: The number of epochs to run.
        max_train_steps_per_epoch: Training will complete after n steps even if loader is not yet exhausted. If None,
            all data will be used.
        max_eval_steps_per_epoch: Evaluation will complete after n steps even if loader is not yet exhausted. If None,
            all data will be used.
        traces: What Traces to run during training. If None, only the system's default Traces will be included.
        log_steps: Frequency (in steps) for printing log messages. 0 to disable all step-based printing (though epoch
            information will still print). None to completely disable printing.
        monitor_names: Additional keys from the data dictionary to be written into the logs.
    """
    monitor_names: Set[str]
    traces_in_use: List[Union[Trace, Scheduler[Trace]]]
    system: System
    filepath: str

    def __init__(self,
                 pipeline: Pipeline,
                 network: BaseNetwork,
                 epochs: int,
                 max_train_steps_per_epoch: Optional[int] = None,
                 max_eval_steps_per_epoch: Optional[int] = None,
                 traces: Union[None, Trace, Scheduler[Trace], Iterable[Union[Trace, Scheduler[Trace]]]] = None,
                 log_steps: Optional[int] = 100,
                 monitor_names: Union[None, str, Iterable[str]] = None):
        self.traces_in_use = []
        self.filepath = os.path.realpath(inspect.stack()[2].filename)  # Record this for history tracking
        assert log_steps is None or log_steps >= 0, \
            "log_steps must be None or positive (or 0 to disable only train logging)"
        self.monitor_names = to_set(monitor_names) | network.get_loss_keys()
        self.system = System(network=network,
                             pipeline=pipeline,
                             traces=to_list(traces),
                             log_steps=log_steps,
                             total_epochs=epochs,
                             max_train_steps_per_epoch=max_train_steps_per_epoch,
                             max_eval_steps_per_epoch=max_eval_steps_per_epoch,
                             system_config=self.fe_summary())

    @property
    def pipeline(self) -> Pipeline:
        return self.system.pipeline

    @property
    def network(self) -> BaseNetwork:
        return self.system.network

    @property
    def traces(self) -> List[Union[Trace, Scheduler[Trace]]]:
        return self.system.traces

    def fit(self, summary: Optional[str] = None, warmup: bool = True, eager: bool = False) -> Optional[Summary]:
        """Train the network for the number of epochs specified by the estimator's constructor.

        Args:
            summary: A name for the experiment. If provided, the log history will be recorded in-memory and returned as
                a summary object at the end of training.
            warmup: Whether to perform warmup before training begins. The warmup procedure will test one step at every
                epoch where schedulers cause the execution graph to change. This can take some time up front, but can
                also save significant heartache on epoch 300 when the training unexpectedly fails due to a tensor size
                mismatch.
            eager: Whether to run the training in eager mode. This is only related to TensorFlow training because
                PyTorch by nature is always in eager mode.

        Returns:
            A summary object containing the training history for this session iff a `summary` name was provided.
        """
        draw()
        self.system.reset(summary, self.fe_summary())
        self._prepare_traces(run_modes={"train", "eval"})
        if warmup:
            self._warmup(eager=eager)
        self._start(run_modes={"train", "eval"}, eager=eager)
        return self.system.summary or None

    def _prepare_traces(self, run_modes: Set[str]) -> None:
        """Prepare information about the traces for execution.

        Add default traces into the traces_in_use list, also prints a warning if no model saver trace is detected.

        Args:
            run_modes: The current execution modes.
        """
        self.traces_in_use = [trace for trace in self.traces]
        if self.system.log_steps is not None:
            self.traces_in_use.append(Logger())
        # Look for any monitor names which should be automagically added.
        trace_outputs = set()
        extra_monitor_keys = set()
        for trace in sort_traces(get_current_items(self.traces_in_use, run_modes=run_modes), ds_ids=[]):
            trace_outputs.update(trace.get_outputs(ds_ids=[]))
            extra_monitor_keys.update(trace.fe_monitor_names - trace_outputs)
        # Add the essential traces
        if "train" in run_modes:
            self.traces_in_use.insert(0, TrainEssential(monitor_names=self.monitor_names.union(extra_monitor_keys)))
            no_save_warning = True
            for trace in get_current_items(self.traces_in_use, run_modes=run_modes):
                if isinstance(trace, (ModelSaver, BestModelSaver)):
                    no_save_warning = False
            if no_save_warning:
                print("FastEstimator-Warn: No ModelSaver Trace detected. Models will not be saved.")
        if "eval" in run_modes and "eval" in self.pipeline.get_modes():
            self.traces_in_use.insert(1, EvalEssential(monitor_names=self.monitor_names.union(extra_monitor_keys)))
        if "test" in run_modes and "test" in self.pipeline.get_modes():
            self.traces_in_use.insert(0, TestEssential(monitor_names=self.monitor_names.union(extra_monitor_keys)))
        # insert system instance to trace
        for trace in get_current_items(self.traces_in_use, run_modes=run_modes):
            trace.system = self.system

    def test(self, summary: Optional[str] = None, eager: bool = False) -> Optional[Summary]:
        """Run the pipeline / network in test mode for one epoch.

        Args:
            summary: A name for the experiment. If provided, the log history will be recorded in-memory and returned as
                a summary object at the end of training. If None, the default value will be whatever `summary` name was
                most recently provided to this Estimator's .fit() or .test() methods.
            eager: Whether to run the training in eager mode. This is only related to TensorFlow training because
                PyTorch by nature is always in eager mode.

        Returns:
            A summary object containing the training history for this session iff the `summary` name is not None (after
            considering the default behavior above).
        """
        self.system.reset_for_test(summary)
        self._prepare_traces(run_modes={"test"})
        self._start(run_modes={"test"}, eager=eager)
        return self.system.summary or None

    def _warmup(self, eager: bool = True) -> None:
        """Perform a test run of each pipeline and network signature epoch to make sure that training won't fail later.

        Traces are not executed in the warmup since they are likely to contain state variables which could become
        corrupted by running extra steps.

        Args:
            eager: Whether to run the training in eager mode. This is only related to TensorFlow training because
                PyTorch by nature is always in eager mode.
        """
        all_traces = get_current_items(self.traces_in_use, run_modes={"train", "eval"})
        sort_traces(all_traces, ds_ids=[])  # This ensures that the traces can sort properly for on_begin and on_end
        monitor_names = self.monitor_names
        for mode in self.pipeline.get_modes() - {"test"}:
            scheduled_items = self.pipeline.get_scheduled_items(mode) + self.network.get_scheduled_items(
                mode) + self.get_scheduled_items(mode)
            signature_epochs = get_signature_epochs(scheduled_items, self.system.total_epochs, mode=mode)
            epochs_with_data = self.pipeline.get_epochs_with_data(total_epochs=self.system.total_epochs, mode=mode)
            for epoch in signature_epochs:
                if epoch not in epochs_with_data:
                    continue
                ds_ids = self.pipeline.get_ds_ids(epoch, mode)
                for ds_id in ds_ids:
                    network_output_keys = self.network.get_all_output_keys(mode, epoch, ds_id=ds_id)
                    network_input_keys = self.network.get_effective_input_keys(mode, epoch, ds_id=ds_id)
                    trace_input_keys = set()
                    trace_output_keys = {"*"}
                    traces = get_current_items(self.traces_in_use, run_modes=mode, epoch=epoch, ds_id=ds_id)
                    for idx, trace in enumerate(traces):
                        if idx > 0:  # ignore TrainEssential and EvalEssential's inputs for unmet requirement checking
                            trace_input_keys.update(trace.inputs)
                        trace_output_keys.update(trace.get_outputs(ds_ids=ds_ids))
                    # key checking
                    with self.pipeline(mode=mode,
                                       epoch=epoch,
                                       ds_id=ds_id,
                                       output_keys=trace_input_keys - network_output_keys
                                       | network_input_keys) as loader:
                        loader = self._configure_loader(loader)
                        with Suppressor():
                            if isinstance(loader, tf.data.Dataset):
                                batch = list(loader.take(1))[0]
                            else:
                                batch = next(iter(loader))
                        batch = self._configure_tensor(loader, batch)
                    assert isinstance(batch, dict), "please make sure data output format is dictionary"
                    pipeline_output_keys = to_set(batch.keys())

                    monitor_names = monitor_names - (pipeline_output_keys | network_output_keys)
                    unmet_requirements = trace_input_keys - (pipeline_output_keys | network_output_keys
                                                             | trace_output_keys)
                    assert not unmet_requirements, \
                        "found missing key(s) during epoch {} mode {} ds_id {}: {}".format(epoch, mode, ds_id,
                                                                                           unmet_requirements)
                    sort_traces(traces, ds_ids=ds_ids, available_outputs=pipeline_output_keys | network_output_keys)
                    trace_input_keys.update(traces[0].inputs)
                    self.network.load_epoch(mode, epoch, ds_id, output_keys=trace_input_keys, warmup=True, eager=eager)
                    self.network.run_step(batch)
                    self.network.unload_epoch()
        assert not monitor_names, "found missing key(s): {}".format(monitor_names)

    def get_scheduled_items(self, mode: str) -> List[Any]:
        """Get a list of items considered for scheduling.

        Args:
            mode: Current execution mode.

        Returns:
            List of schedulable items in estimator.
        """
        return self.traces_in_use

    def _start(self, run_modes: Set[str], eager: bool) -> None:
        """The outer training loop.

        This method invokes the trace on_begin method, runs the necessary 'train' and 'eval' epochs, and then invokes
        the trace on_end method.

        Args:
            run_modes: The current execution modes.
            eager: Whether to run the training in eager mode. This is only related to TensorFlow training because
                PyTorch by nature is always in eager mode.
        """
        all_traces = sort_traces(get_current_items(self.traces_in_use, run_modes=run_modes), ds_ids=[])
        with NonContext() if fe.fe_history_path is False else HistoryRecorder(
                self.system, self.filepath, db_path=fe.fe_history_path):
            try:
                self._run_traces_on_begin(traces=all_traces)
                if "train" in run_modes or "eval" in run_modes:
                    # If the training is re-starting from a restore wizard, it should re-run the last eval epoch
                    if self.system.epoch_idx > 0 and "eval" in self.pipeline.get_modes(epoch=self.system.epoch_idx):
                        self.system.mode = "eval"
                        self._run_epoch(eager=eager)
                    for self.system.epoch_idx in range(self.system.epoch_idx + 1, self.system.total_epochs + 1):
                        if "train" in self.pipeline.get_modes(epoch=self.system.epoch_idx):
                            self.system.mode = "train"
                            self._run_epoch(eager=eager)
                        if "eval" in self.pipeline.get_modes(epoch=self.system.epoch_idx):
                            self.system.mode = "eval"
                            self._run_epoch(eager=eager)
                else:
                    self._run_epoch(eager=eager)
            except EarlyStop:
                pass  # On early stopping we still want to run the final traces and return results
            self._run_traces_on_end(traces=all_traces)

    def _run_epoch(self, eager: bool) -> None:
        """A method to perform an epoch of activity.

        This method requires that the current mode and epoch already be specified within the self.system object.

        Args:
            eager: Whether to run the training in eager mode. This is only related to TensorFlow training because
                PyTorch by nature is always in eager mode.
        """
        ds_ids = self.pipeline.get_ds_ids(self.system.epoch_idx, self.system.mode)
        epoch_traces = sort_traces(
            get_current_items(self.traces_in_use, run_modes=self.system.mode, epoch=self.system.epoch_idx),
            ds_ids=ds_ids)
        self._run_traces_on_epoch_begin(traces=epoch_traces)
        self.system.batch_idx = None
        end_epoch_data = Data()  # We will aggregate data over on_ds_end and put it into on_epoch_end for printing
        # run for each dataset
        for self.system.ds_id in ds_ids:
            ds_traces = get_current_items(self.traces_in_use,
                                          run_modes=self.system.mode,
                                          epoch=self.system.epoch_idx,
                                          ds_id=self.system.ds_id)
            trace_input_keys = set()
            for ds_trace in ds_traces:
                trace_input_keys.update(ds_trace.inputs)
            network_input_keys = self.network.get_effective_input_keys(mode=self.system.mode,
                                                                       epoch=self.system.epoch_idx,
                                                                       ds_id=self.system.ds_id)
            network_output_keys = self.network.get_all_output_keys(mode=self.system.mode,
                                                                   epoch=self.system.epoch_idx,
                                                                   ds_id=self.system.ds_id)
            self.network.load_epoch(mode=self.system.mode,
                                    epoch=self.system.epoch_idx,
                                    ds_id=self.system.ds_id,
                                    output_keys=trace_input_keys,
                                    eager=eager)
            with self.pipeline(mode=self.system.mode,
                               epoch=self.system.epoch_idx,
                               ds_id=self.system.ds_id,
                               output_keys=trace_input_keys - network_output_keys | network_input_keys) as loader:
                loader = self._configure_loader(loader)
                iterator = iter(loader)
                with Suppressor():
                    batch = next(iterator)
                ds_traces = sort_traces(ds_traces,
                                        available_outputs=to_set(batch.keys()) | network_output_keys,
                                        ds_ids=ds_ids)
                per_ds_traces = [trace for trace in ds_traces if isinstance(trace, PerDSTrace)]
                self._run_traces_on_ds_begin(traces=per_ds_traces)
                while True:
                    try:
                        if self.system.mode == "train":
                            self.system.update_global_step()
                        self.system.update_batch_idx()
                        batch = self._configure_tensor(loader, batch)
                        self._run_traces_on_batch_begin(batch, traces=ds_traces)
                        batch, prediction = self.network.run_step(batch)
                        self._run_traces_on_batch_end(batch, prediction, traces=ds_traces)
                        if isinstance(loader,
                                      DataLoader) and ((self.system.batch_idx == self.system.max_train_steps_per_epoch
                                                        and self.system.mode == "train") or
                                                       (self.system.batch_idx == self.system.max_eval_steps_per_epoch
                                                        and self.system.mode == "eval")):
                            raise StopIteration
                        with Suppressor():
                            batch = next(iterator)
                    except StopIteration:
                        break
                self._run_traces_on_ds_end(traces=per_ds_traces, data=end_epoch_data)
            self.network.unload_epoch()
        self._run_traces_on_epoch_end(traces=epoch_traces, data=end_epoch_data)

    def _configure_loader(self, loader: Union[DataLoader, tf.data.Dataset]) -> Union[DataLoader, tf.data.Dataset]:
        """A method to configure a given dataloader for use with this Estimator's Network.

        This method will ensure that the `loader` returns the correct data type (tf.Tensor or torch.Tensor) depending on
         the requirements of the Network. It also handles issues with multi-gpu data sharding.

        Args:
            loader: A data loader to be modified.

        Returns:
            The potentially modified dataloader to be used for training.
        """
        new_loader = loader
        if isinstance(new_loader, DataLoader) and isinstance(self.network, TFNetwork):
            add_batch = True
            if isinstance(loader.dataset, OpDataset) and loader.dataset.dataset.fe_batch:
                add_batch = False
            batch = to_tensor(loader.dataset[0], target_type="tf")
            data_type = to_type(batch)
            data_shape = to_shape(batch, add_batch=add_batch, exact_shape=False)
            new_loader = tf.data.Dataset.from_generator(lambda: loader, data_type, output_shapes=data_shape)
            new_loader = new_loader.prefetch(1)
        if isinstance(new_loader, tf.data.Dataset):
            if self.system.max_train_steps_per_epoch and self.system.mode == "train":
                new_loader = new_loader.take(self.system.max_train_steps_per_epoch)
            if self.system.max_eval_steps_per_epoch and self.system.mode == "eval":
                new_loader = new_loader.take(self.system.max_eval_steps_per_epoch)
            if isinstance(tf.distribute.get_strategy(), tf.distribute.MirroredStrategy) and isinstance(
                    self.network, TFNetwork) and not isinstance(new_loader, DistributedDataset):
                # The default autoshard policy is file, changing it to data to avoid warning
                options = tf.data.Options()
                options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
                new_loader = new_loader.with_options(options)
                new_loader = tf.distribute.get_strategy().experimental_distribute_dataset(new_loader)
        return new_loader

    def _configure_tensor(self, loader: Union[DataLoader, tf.data.Dataset], batch: Dict[str, Any]) -> Dict[str, Any]:
        """A function to convert a batch of tf.Tensors to torch.Tensors if required.

        Returns:
            Either the original `batch`, or the `batch` converted to torch.Tensors if required.
        """
        if isinstance(loader, tf.data.Dataset) and isinstance(self.network, TorchNetwork):
            batch = to_tensor(batch, target_type="torch")
        return batch

    def _run_traces_on_begin(self, traces: Iterable[Trace]) -> None:
        """Invoke the on_begin methods of given traces.

        Args:
            traces: List of traces.
        """
        data = Data()
        restore = None
        for trace in traces:
            # Delay RestoreWizard until the end so that it can overwrite everyone's on_begin methods
            if isinstance(trace, RestoreWizard):
                restore = trace
                continue
            # Restore does need to run before the logger though
            if isinstance(trace, Logger) and restore:
                restore.on_begin(data)
                restore = None
            trace.on_begin(data)
        if restore:
            restore.on_begin(data)
        self._check_early_exit()

    def _run_traces_on_epoch_begin(self, traces: Iterable[Trace]) -> None:
        """Invoke the on_epoch_begin methods of given traces.

        Args:
            traces: List of traces.
        """
        data = Data()
        for trace in traces:
            trace.on_epoch_begin(data)
        self._check_early_exit()

    def _run_traces_on_ds_begin(self, traces: Iterable[PerDSTrace]) -> None:
        """Invoke the on_ds_begin methods of given traces.

        Args:
            traces: List of traces.
        """
        data = Data()
        for trace in traces:
            trace.on_ds_begin(data)
        self._check_early_exit()

    def _run_traces_on_batch_begin(self, batch: Dict[str, Any], traces: Iterable[Trace]) -> None:
        """Invoke the on_batch_begin methods of given traces.

        Args:
            batch: The batch data which was provided by the pipeline.
            traces: List of traces.
        """
        data = Data(batch)
        for trace in traces:
            trace.on_batch_begin(data)
        self._check_early_exit()

    def _run_traces_on_batch_end(self, batch: Dict[str, Any], prediction: Dict[str, Any],
                                 traces: Iterable[Trace]) -> None:
        """Invoke the on_batch_end methods of given traces.

        Args:
            batch: The batch data which was provided by the pipeline.
            prediction: The prediction data which was generated by the network.
            traces: List of traces.
        """
        data = Data(ChainMap(prediction, batch))
        for trace in traces:
            trace.on_batch_end(data)
        self._check_early_exit()

    def _run_traces_on_ds_end(self, traces: Iterable[PerDSTrace], data: Data) -> None:
        """Invoke the on_ds_begin methods of given traces.

        Args:
            traces: List of traces.
            data: Data into which to record results.
        """
        for trace in traces:
            trace.on_ds_end(data)
        self._check_early_exit()

    def _run_traces_on_epoch_end(self, traces: Iterable[Trace], data: Data) -> None:
        """Invoke the on_epoch_end methods of of given traces.

        Args:
            traces: List of traces.
            data: Data into which to record results.
        """
        for trace in traces:
            trace.on_epoch_end(data)
        self._check_early_exit()

    @staticmethod
    def _run_traces_on_end(traces: Iterable[Trace]) -> None:
        """Invoke the on_end methods of given traces.

        Args:
            traces: List of traces.
        """
        data = Data()
        traceability = None
        for trace in traces:
            if isinstance(trace, Traceability):
                # Delay traceability until the end so that it can capture all data including the total training time
                traceability = trace
                continue
            trace.on_end(data)
        if traceability:
            traceability.on_end(data)

    def _check_early_exit(self) -> None:
        """Determine whether training should be prematurely aborted.

        Raises:
            EarlyStop: If the system.stop_training flag has been set to True.
        """
        if self.system.stop_training:
            raise EarlyStop


class EarlyStop(Exception):
    """An exception raised when the system.stop_training flag is flipped by a Trace in order to abort the training.

    This class is intentionally not @traceable.
    """


def enable_deterministic(seed: int) -> None:
    """Invoke to set random seed for deterministic training.

    The determinism only works for tensorflow >= 2.1 and pytorch >= 1.14, and some model layers don't support.

    Known failing layers:
    * tf.keras.layers.UpSampling2D

    Args:
        seed: The random seed to use for training.
    """
    fe.fe_deterministic_seed = seed
    os.environ['PYTHONHASHSEED'] = str(seed)
    os.environ['TF_DETERMINISTIC_OPS'] = str(1)
    random.seed(seed)
    np.random.seed(seed)
    tf.random.set_seed(seed)
    torch.manual_seed(seed)


def record_history(path: Union[bool, str]) -> None:
    """Change the default location for history tracking.

    Args:
        path: The path to save experiment histories. Pass True to use the default location of
            ~/fastestimator_data/history.db. Pass False to disable history tracking.
    """
    if path in (None, True):
        fe.fe_history_path = None
    else:
        fe.fe_history_path = path



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import inspect
import json
import os
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Tuple

from fastestimator.network import build


class Search(ABC):
    """Base class which other searches inherit from.

    The base search class takes care of evaluation logging, saving and loading, and is also able to recover from
    interrupted search runs and cache the search history.

    Args:
        score_fn: Objective function that measures search fitness. One of its arguments must be 'search_idx' which will
            be automatically provided by the search routine. This can help with file saving / logging during the search.
        best_mode: Whether maximal or minimal fitness is desired. Must be either 'min' or 'max'.
        name: The name of the search instance. This is used for saving and loading purposes.

    Raises:
        AssertionError: If `best_mode` is not 'min' or 'max', or search_idx is not an input argument of `score_fn`.
    """
    def __init__(self, score_fn: Callable[..., float], best_mode: str = "max", name: str = "search"):
        assert best_mode in ["max", "min"], "best_mode must be either 'max' or 'min'"
        assert "search_idx" in inspect.signature(score_fn).parameters, \
            "score_fn must take 'search_idx' as one of its input arguments"
        self.score_fn = score_fn
        self.best_mode = best_mode
        self.name = name
        self.save_dir = None
        self._initialize_state()

    def _initialize_state(self):
        self.search_idx = 0
        self.search_results = []
        self.evaluation_cache = {}

    def evaluate(self, **kwargs: Any) -> float:
        """Evaluate the score function and return the score.

        Args:
            kwargs: Any keyword argument(s) to pass to the score function. Should not contain search_idx as this will be
                populated manually here.

        Returns:
            Fitness score calculated by `score_fn`.
        """
        # evaluation caching
        hash_value = hash(tuple(sorted(kwargs.items())))
        if hash_value in self.evaluation_cache:
            score = self.evaluation_cache[hash_value]
        else:
            self.search_idx += 1
            build.count = 0  # Resetting the build count to refresh the model names
            kwargs["search_idx"] = self.search_idx
            score = self.score_fn(**kwargs)
            self.search_results.append((kwargs, score))
            self.evaluation_cache[hash_value] = score
            if self.save_dir is not None:
                self.save(self.save_dir)
            print("FastEstimator-Search: Evaluated {}, score: {}".format(kwargs, score))
        return score

    def get_best_results(self) -> Tuple[Dict[str, Any], float]:
        """Get the best result from the current search history.

        Returns:
            The best results in the format of (parameter, score)

        Raises:
            RuntimeError: If the search hasn't been run yet.
        """
        if not self.search_results:
            raise RuntimeError("No search has been run yet, so best parameters are not available.")
        if self.best_mode == "max":
            best_results = max(self.search_results, key=lambda x: x[1])
        else:  # min
            best_results = min(self.search_results, key=lambda x: x[1])
        return best_results

    def get_search_results(self) -> List[Tuple[Dict[str, Any], float]]:
        """Get the current search history.

        Returns:
            The evluation history list, with each element being a tuple of parameters and score.
        """
        return self.search_results.copy()

    def _get_state(self) -> Dict[str, Any]:
        """Get the current state of the search instance, the state is the variables that can be saved or loaded.

        Returns:
            The dictionary containing the state variable.
        """
        return {"search_results": self.search_results}

    def save(self, save_dir: str) -> None:
        """Save the state of the instance to a specific directory, it will create `name.json` file in the `save_dir`.

        Args:
            save_dir: The folder path to save to.
        """
        file_path = os.path.join(save_dir, "{}.json".format(self.name))
        with open(file_path, 'w') as fp:
            json.dump(self._get_state(), fp, indent=4)
        print("FastEstimator-Search: Saving the search state to {}".format(file_path))

    def load(self, load_dir: str, not_exist_ok: bool = False) -> None:
        """Load the state of search from a given directory. It will look for `name.json` within the `load_dir`.

        Args:
            load_dir: The folder path to load the state from.
            not_exist_ok: whether to ignore when the file does not exist.
        """
        self._initialize_state()
        # load from file
        load_dir = os.path.abspath(os.path.normpath(load_dir))
        file_path = os.path.join(load_dir, "{}.json".format(self.name))
        if os.path.exists(file_path):
            with open(file_path, 'r') as fp:
                state = json.load(fp)
            # restore all state variables from get_state
            self.__dict__.update(state)
            # restore evaluation cache and search_idx
            for kwargs, score in self.search_results:
                kwargs = kwargs.copy()
                search_idx = kwargs.pop('search_idx')  # This won't appear in the hash later
                self.search_idx = self.search_idx if self.search_idx > search_idx else search_idx
                # Each python session uses a unique salt for hash, so can't save the hashes to disk for re-use
                self.evaluation_cache[hash(tuple(sorted(kwargs.items())))] = score
            print("FastEstimator-Search: Loading the search state from {}".format(file_path))
        elif not not_exist_ok:
            raise ValueError("cannot find file to load in {}".format(file_path))

    def fit(self, save_dir: str = None) -> None:
        """Start the search.

        Args:
            save_dir: When `save_dir` is provided, the search results will be backed up to the `save_dir` after each
                evaluation. It will also attempt to load the search state from `save_dir` if possible. This is useful
                when the search might experience interruption since it can be restarted using the same command.
        """
        if save_dir is None:
            self._initialize_state()
        else:
            self.save_dir = save_dir
            os.makedirs(save_dir, exist_ok=True)
            self.load(save_dir, not_exist_ok=True)
        self._fit()

    @abstractmethod
    def _fit(self) -> None:
        raise NotImplementedError



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, Optional, TypeVar, Union

import tensorflow as tf
import torch

from fastestimator.util.util import NonContext

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


def get_gradient(target: Tensor,
                 sources: Union[Iterable[Tensor], Tensor],
                 higher_order: bool = False,
                 tape: Optional[tf.GradientTape] = None,
                 retain_graph: bool = True) -> Union[Iterable[Tensor], Tensor]:
    """Calculate gradients of a target w.r.t sources.

    This method can be used with TensorFlow tensors:
    ```python
    x = tf.Variable([1.0, 2.0, 3.0])
    with tf.GradientTape(persistent=True) as tape:
        y = x * x

        b = fe.backend.get_gradient(target=y, sources=x, tape=tape)  # [2.0, 4.0, 6.0]
        b = fe.backend.get_gradient(target=b, sources=x, tape=tape)  # None

        b = fe.backend.get_gradient(target=y, sources=x, tape=tape, higher_order=True)  # [2.0, 4.0, 6.0]
        b = fe.backend.get_gradient(target=b, sources=x, tape=tape)  # [2.0, 2.0, 2.0]
    ```

    This method can be used with PyTorch tensors:
    ```python
    x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
    y = x * x

    b = fe.backend.get_gradient(target=y, sources=x)  # [2.0, 4.0, 6.0]
    b = fe.backend.get_gradient(target=b, sources=x)  # Error - b does not have a backwards function

    b = fe.backend.get_gradient(target=y, sources=x, higher_order=True)  # [2.0, 4.0, 6.0]
    b = fe.backend.get_gradient(target=b, sources=x)  # [2.0, 2.0, 2.0]
    ```

    Args:
        target: The target (final) tensor.
        sources: A sequence of source (initial) tensors.
        higher_order: Whether the gradient will be used for higher order gradients.
        tape: TensorFlow gradient tape. Only needed when using the TensorFlow backend.
        retain_graph: Whether to retain PyTorch graph. Only valid when using the PyTorch backend.

    Returns:
        Gradient(s) of the `target` with respect to the `sources`.

    Raises:
        ValueError: If `target` is an unacceptable data type.
    """
    if tf.is_tensor(target):
        with NonContext() if higher_order else tape.stop_recording():
            gradients = tape.gradient(target, sources)
    elif isinstance(target, torch.Tensor):
        gradients = torch.autograd.grad(target,
                                        sources,
                                        grad_outputs=torch.ones_like(target),
                                        retain_graph=retain_graph,
                                        create_graph=higher_order,
                                        only_inputs=True)

        if isinstance(sources, torch.Tensor):
            #  The behavior table of tf and torch backend
            #  ---------------------------------------------------------------
            #        | case 1                     | case 2                    |
            #  ---------------------------------------------------------------|
            #  tf    | target: tf.Tensor          | target: tf.Tensor         |
            #        | sources: tf.Tensor         | sources: [tf.Tensor]      |
            #        | gradients: tf.Tensor       | gradients: [tf.Tensor]    |
            # ----------------------------------------------------------------|
            #  torch | target: torch.Tensor       | target: tf.Tensor         |
            #        | sources: torch.Tensor      | sources: [tf.Tensor]      |
            #        | gradients: (torch.Tensor,) | gradients: (torch.Tensor,)|
            # ----------------------------------------------------------------
            # In order to make the torch behavior become the same as tf in case 1, need to unwrap the gradients when
            # source is not Iterable.

            gradients = gradients[0]
    else:
        raise ValueError("Unrecognized tensor type {}".format(type(target)))
    return gradients



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, Optional, Tuple, Union

from albumentations.augmentations.transforms import MaskDropout as MaskDropoutAlb

from fastestimator.op.numpyop.multivariate.multivariate import MultiVariateAlbumentation
from fastestimator.util.traceability_util import traceable


@traceable()
class MaskDropout(MultiVariateAlbumentation):
    """Zero out objects from an image + mask pair.

    An image & mask augmentation that zero out mask and image regions corresponding to randomly chosen object instance
    from mask. The mask must be single-channel image, with zero values treated as background. The image can be any
    number of channels.

    Args:
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        image_in: The key of an image to be modified.
        mask_in: The key of a mask to be modified (with the same random factors as the image).
        masks_in: The key of masks to be modified (with the same random factors as the image).
                image_out: The key to write the modified image (defaults to `image_in` if None).
        mask_out: The key to write the modified mask (defaults to `mask_in` if None).
        masks_out: The key to write the modified masks (defaults to `masks_in` if None).
        max_objects: Maximum number of labels that can be zeroed out. Can be tuple, in this case it's [min, max]
        image_fill_value: Fill value to use when filling image.
            Can be 'inpaint' to apply in-painting (works only  for 3-channel images)
        mask_fill_value: Fill value to use when filling mask.

    Image types:
        uint8, float32
    """
    def __init__(self,
                 max_objects: Union[int, Tuple[int, int]] = 1,
                 image_fill_value: Union[int, float, str] = 0,
                 mask_fill_value: Union[int, float] = 0,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 image_in: Optional[str] = None,
                 mask_in: Optional[str] = None,
                 masks_in: Optional[str] = None,
                 image_out: Optional[str] = None,
                 mask_out: Optional[str] = None,
                 masks_out: Optional[str] = None):
        super().__init__(
            MaskDropoutAlb(max_objects=max_objects,
                           image_fill_value=image_fill_value,
                           mask_fill_value=mask_fill_value,
                           always_apply=True),
            image_in=image_in,
            mask_in=mask_in,
            masks_in=masks_in,
            bbox_in=None,
            keypoints_in=None,
            image_out=image_out,
            mask_out=mask_out,
            masks_out=masks_out,
            bbox_out=None,
            keypoints_out=None,
            bbox_params=None,
            keypoint_params=None,
            mode=mode,
            ds_id=ds_id)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, Iterable, List, TypeVar, Union

import tensorflow as tf
import torch

from fastestimator.backend.gather_from_batch import gather_from_batch
from fastestimator.backend.reduce_max import reduce_max
from fastestimator.op.tensorop.tensorop import TensorOp
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_list

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


@traceable()
class Gather(TensorOp):
    """Gather values from an input tensor.

    If indices are not provided, the maximum values along the batch dimension will be collected.

    Args:
        inputs: The tensor(s) to gather values from.
        indices: A tensor containing target indices to gather.
        outputs: The key(s) under which to save the output.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 inputs: Union[str, List[str]],
                 outputs: Union[str, List[str]],
                 indices: Union[None, str, List[str]] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None):
        indices = to_list(indices)
        self.num_indices = len(indices)
        combined_inputs = indices
        combined_inputs.extend(to_list(inputs))
        super().__init__(inputs=combined_inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.in_list, self.out_list = True, True

    def forward(self, data: List[Tensor], state: Dict[str, Any]) -> List[Tensor]:
        indices = data[:self.num_indices]
        inputs = data[self.num_indices:]
        results = []
        for idx, tensor in enumerate(inputs):
            # Check len(indices[0]) since an empty indices element is used to trigger the else
            if tf.is_tensor(indices[0]) or isinstance(indices[0], torch.Tensor):
                elem_len = indices[0].shape[0]
            else:
                elem_len = len(indices[0])
            if len(indices) > idx and elem_len > 0:
                results.append(gather_from_batch(tensor, indices=indices[idx]))
            elif len(indices) == 1 and elem_len > 0:
                # One set of indices for all outputs
                results.append(gather_from_batch(tensor, indices=indices[0]))
            else:
                results.append(reduce_max(tensor, 1))  # The maximum value within each batch element
        return results



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, List, Optional, Union

from albumentations import BboxParams, KeypointParams
from albumentations.augmentations.transforms import CropNonEmptyMaskIfExists as CropNonEmptyMaskIfExistsAlb

from fastestimator.op.numpyop.multivariate.multivariate import MultiVariateAlbumentation
from fastestimator.util.traceability_util import traceable


@traceable()
class CropNonEmptyMaskIfExists(MultiVariateAlbumentation):
    """Crop an area with mask if mask is non-empty, otherwise crop randomly.

    Args:
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        image_in: The key of an image to be modified.
        mask_in: The key of a mask to be modified (with the same random factors as the image).
        masks_in: The key of masks to be modified (with the same random factors as the image).
        bbox_in: The key of a bounding box(es) to be modified (with the same random factors as the image).
        keypoints_in: The key of keypoints to be modified (with the same random factors as the image).
        image_out: The key to write the modified image (defaults to `image_in` if None).
        mask_out: The key to write the modified mask (defaults to `mask_in` if None).
        masks_out: The key to write the modified masks (defaults to `masks_in` if None).
        bbox_out: The key to write the modified bounding box(es) (defaults to `bbox_in` if None).
        keypoints_out: The key to write the modified keypoints (defaults to `keypoints_in` if None).
        bbox_params: Parameters defining the type of bounding box ('coco', 'pascal_voc', 'albumentations' or 'yolo').
        keypoint_params: Parameters defining the type of keypoints ('xy', 'yx', 'xya', 'xys', 'xyas', 'xysa').
        height: Vertical size of crop in pixels.
        width: Horizontal size of crop in pixels.
        ignore_values: Values to ignore in mask, `0` values are always ignored (e.g. if background value is 5 set
            `ignore_values=[5]` to ignore).
        ignore_channels: Channels to ignore in mask (e.g. if background is a first channel set `ignore_channels=[0]` to
            ignore).

    Image types:
        uint8, float32
    """
    def __init__(self,
                 height: int,
                 width: int,
                 ignore_values: Optional[List[int]] = None,
                 ignore_channels: Optional[List[int]] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 image_in: Optional[str] = None,
                 mask_in: Optional[str] = None,
                 masks_in: Optional[str] = None,
                 bbox_in: Optional[str] = None,
                 keypoints_in: Optional[str] = None,
                 image_out: Optional[str] = None,
                 mask_out: Optional[str] = None,
                 masks_out: Optional[str] = None,
                 bbox_out: Optional[str] = None,
                 keypoints_out: Optional[str] = None,
                 bbox_params: Union[BboxParams, str, None] = None,
                 keypoint_params: Union[KeypointParams, str, None] = None):
        super().__init__(
            CropNonEmptyMaskIfExistsAlb(height=height,
                                        width=width,
                                        ignore_values=ignore_values,
                                        ignore_channels=ignore_channels,
                                        always_apply=True),
            image_in=image_in,
            mask_in=mask_in,
            masks_in=masks_in,
            bbox_in=bbox_in,
            keypoints_in=keypoints_in,
            image_out=image_out,
            mask_out=mask_out,
            masks_out=masks_out,
            bbox_out=bbox_out,
            keypoints_out=keypoints_out,
            bbox_params=bbox_params,
            keypoint_params=keypoint_params,
            mode=mode,
            ds_id=ds_id)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tempfile
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.op.tensorop.loss import CrossEntropy, Hinge, SuperLoss
from fastestimator.op.tensorop.model import ModelOp
from fastestimator.test.unittest_util import sample_system_object


class TestSuperLoss(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # torch binary ce
        cls.torch_true_binary = torch.tensor([[1], [0], [1], [0]]).to("cuda:0" if torch.cuda.is_available() else "cpu")
        cls.torch_pred_binary = torch.tensor([[0.9], [0.3], [0.8],
                                              [0.1]]).to("cuda:0" if torch.cuda.is_available() else "cpu")
        # categorical ce
        cls.tf_true_cat = tf.constant([[0, 1, 0], [1, 0, 0], [0, 0, 1]])
        cls.tf_pred_cat = tf.constant([[0.1, 0.8, 0.1], [0.9, 0.05, 0.05], [0.1, 0.2, 0.7]])
        cls.torch_true_cat = torch.tensor([[0, 1, 0], [1, 0, 0],
                                           [0, 0, 1]]).to("cuda:0" if torch.cuda.is_available() else "cpu")
        cls.torch_pred_cat = torch.tensor([[0.1, 0.8, 0.1], [0.9, 0.05, 0.05],
                                           [0.1, 0.2, 0.7]]).to("cuda:0" if torch.cuda.is_available() else "cpu")
        # sparse categorical ce
        cls.tf_true_sparse = tf.constant([[0], [1], [0]])
        cls.tf_pred_sparse = tf.constant([[0.1, 0.8, 0.1], [0.9, 0.05, 0.05], [0.1, 0.2, 0.7]])
        cls.state = {'warmup': False, 'epoch': 1, 'mode': 'train'}

    @staticmethod
    @tf.function
    def do_forward(op, data, state):
        return op.forward(data, state)

    def test_tf_superloss_categorical_ce(self):
        sl = SuperLoss(CrossEntropy(inputs=['y_pred', 'y'], outputs='ce'))
        sl.build(framework="tf", device=None)
        output = sl.forward(data=[self.tf_pred_cat, self.tf_true_cat], state=self.state)
        self.assertTrue(np.allclose(output.numpy(), -0.0026386082))

    def test_tf_static_superloss_categorical_ce(self):
        sl = SuperLoss(CrossEntropy(inputs=['y_pred', 'y'], outputs='ce'))
        sl.build(framework="tf", device=None)
        output = self.do_forward(sl, data=[self.tf_pred_cat, self.tf_true_cat], state=self.state)
        self.assertTrue(np.allclose(output.numpy(), -0.0026386082))

    def test_tf_superloss_sparse_categorical_ce(self):
        sl = SuperLoss(CrossEntropy(inputs=['y_pred', 'y'], outputs='ce'))
        sl.build(framework="tf", device=None)
        output = sl.forward(data=[self.tf_pred_sparse, self.tf_true_sparse], state=self.state)
        self.assertTrue(np.allclose(output.numpy(), -0.024740249))

    def test_tf_static_superloss_sparse_categorical_ce(self):
        sl = SuperLoss(CrossEntropy(inputs=['y_pred', 'y'], outputs='ce'))
        sl.build(framework="tf", device=None)
        output = self.do_forward(sl, data=[self.tf_pred_sparse, self.tf_true_sparse], state=self.state)
        self.assertTrue(np.allclose(output.numpy(), -0.024740249))

    def test_torch_superloss_binary_ce(self):
        sl = SuperLoss(CrossEntropy(inputs=['y_pred', 'y'], outputs='ce'))
        sl.build(framework="torch", device="cuda:0" if torch.cuda.is_available() else "cpu")
        output = sl.forward(data=[self.torch_pred_binary, self.torch_true_binary], state=self.state)
        self.assertTrue(np.allclose(output.detach().to("cpu").numpy(), -0.0026238672))

    def test_tf_superloss_hinge(self):
        true = tf.constant([[-1, 1, 1, -1], [1, 1, 1, 1], [-1, -1, 1, -1], [1, -1, -1, -1]])
        pred = tf.constant([[0.1, 0.9, 0.05, 0.05], [0.1, -0.2, 0.0, -0.7], [0.0, 0.15, 0.8, 0.05],
                            [1.0, -1.0, -1.0, -1.0]])
        sl = SuperLoss(Hinge(inputs=('x1', 'x2'), outputs='x'))
        sl.build('tf', None)
        output = sl.forward(data=[pred, true], state=self.state)
        self.assertTrue(np.allclose(output.numpy(), -0.072016776))

    def test_tf_static_superloss_hinge(self):
        true = tf.constant([[-1, 1, 1, -1], [1, 1, 1, 1], [-1, -1, 1, -1], [1, -1, -1, -1]])
        pred = tf.constant([[0.1, 0.9, 0.05, 0.05], [0.1, -0.2, 0.0, -0.7], [0.0, 0.15, 0.8, 0.05],
                            [1.0, -1.0, -1.0, -1.0]])
        sl = SuperLoss(Hinge(inputs=('x1', 'x2'), outputs='x'))
        sl.build('tf', None)
        output = self.do_forward(sl, data=[pred, true], state=self.state)
        self.assertTrue(np.allclose(output.numpy(), -0.072016776))

    def test_torch_superloss_hinge(self):
        true = torch.tensor([[-1, 1, 1, -1], [1, 1, 1, 1], [-1, -1, 1, -1],
                             [1, -1, -1, -1]]).to("cuda:0" if torch.cuda.is_available() else "cpu")
        pred = torch.tensor([[0.1, 0.9, 0.05, 0.05], [0.1, -0.2, 0.0, -0.7], [0.0, 0.15, 0.8, 0.05],
                             [1.0, -1.0, -1.0, -1.0]]).to("cuda:0" if torch.cuda.is_available() else "cpu")
        sl = SuperLoss(Hinge(inputs=('x1', 'x2'), outputs='x'))
        sl.build('torch', "cuda:0" if torch.cuda.is_available() else "cpu")
        output = sl.forward(data=[pred, true], state=self.state)
        self.assertTrue(np.allclose(output.to("cpu").numpy(), -0.072016776))

    def test_save_and_load_state_tf(self):
        def instantiate_system():
            system = sample_system_object()
            model = fe.build(model_fn=fe.architecture.tensorflow.LeNet, optimizer_fn='adam', model_name='tf')
            system.network = fe.Network(ops=[
                ModelOp(model=model, inputs="x_out", outputs="y_pred"),
                SuperLoss(CrossEntropy(inputs=['y_pred', 'y'], outputs='ce'))
            ])
            return system

        system = instantiate_system()

        # make some changes
        system.network.ops[1].forward(data=[self.tf_pred_cat, self.tf_true_cat], state=self.state)

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # re-instantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)

        loaded_op = system.network.ops[1]
        with self.subTest("Initialization Flags"):
            self.assertEqual(loaded_op.initialized['train'].numpy(), True)
            self.assertEqual(loaded_op.initialized['eval'].numpy(), False)
        with self.subTest("Mean Values"):
            self.assertTrue(np.allclose(loaded_op.tau['train'].numpy(), 0.22839302))
            self.assertTrue(np.allclose(loaded_op.tau['eval'].numpy(), 0.0))

    def test_save_and_load_state_torch(self):
        def instantiate_system():
            system = sample_system_object()
            model = fe.build(model_fn=fe.architecture.pytorch.LeNet, optimizer_fn='adam', model_name='tf')
            system.network = fe.Network(ops=[
                ModelOp(model=model, inputs="x_out", outputs="y_pred"),
                SuperLoss(CrossEntropy(inputs=['y_pred', 'y'], outputs='ce'))
            ])
            return system

        system = instantiate_system()
        # make some changes
        system.network.ops[1].forward(data=[self.torch_pred_cat, self.torch_true_cat], state=self.state)

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # re-instantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)

        loaded_op = system.network.ops[1]
        with self.subTest("Initialization Flags"):
            self.assertEqual(loaded_op.initialized['train'].to("cpu").numpy(), True)
            self.assertEqual(loaded_op.initialized['eval'].to("cpu").numpy(), False)
        with self.subTest("Mean Values"):
            self.assertTrue(np.allclose(loaded_op.tau['train'].to("cpu").numpy(), 0.22839302))
            self.assertTrue(np.allclose(loaded_op.tau['eval'].to("cpu").numpy(), 0.0))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from fastestimator.util.data import Data
from fastestimator.util.img_data import ImgData
from fastestimator.util.latex_util import AdjustBox, Center, ContainerList, HrefFEID, PyContainer, Verbatim
from fastestimator.util.traceability_util import FeSplitSummary, trace_model, traceable
from fastestimator.util.util import FEID, DefaultKeyDict, Flag, LogSplicer, NonContext, Suppressor, Timer, \
    check_ds_id, check_io_names, draw, get_batch_size, get_num_devices, get_shape, get_type, is_number, pad_batch, \
    pad_data, param_to_range, parse_modes, parse_string_to_python, prettify_metric_name, show_image, strip_prefix, \
    strip_suffix, to_list, to_number, to_set
from fastestimator.util.wget_util import bar_custom, callback_progress



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tempfile
import unittest

import dill
import numpy as np

from fastestimator.op.numpyop.univariate import Calibrate
from fastestimator.test.unittest_util import is_equal


class TestCalibrate(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.single_input = [np.array([1, 2, 3, 5])]
        cls.single_output = [np.array([0.5, 1.0, 1.5, 2.5])]
        cls.multi_input = [np.array([2, 2]), np.array([0, 1, 2])]
        cls.multi_output = [np.array([1, 1]), np.array([0, 0.5, 1])]

    def test_single_input(self):
        op = Calibrate(inputs='x', outputs='x', calibration_fn=lambda x: x/2)
        data = op.forward(data=self.single_input, state={})
        self.assertTrue(is_equal(data, self.single_output))

    def test_multi_input(self):
        op = Calibrate(inputs=['x', 'y'], outputs=['x', 'y'], calibration_fn=lambda x: x/2)
        data = op.forward(data=self.multi_input, state={})
        self.assertTrue(is_equal(data, self.multi_output))

    def test_single_input_fn_from_disk(self):
        tmpdirname = tempfile.mkdtemp()
        fn_path = os.path.join(tmpdirname, 'fn.pkl')
        fn = lambda x: x/2
        with open(fn_path, 'wb') as f:
            dill.dump(fn, f)
        op = Calibrate(inputs='x', outputs='x', calibration_fn=fn_path)
        with self.subTest("Do nothing on Warmup"):
            resp = op.forward(self.single_input, state={'warmup': True})
            self.assertTrue(np.array_equal(resp, self.single_input))
        with self.subTest("Load function during regular execution"):
            resp = op.forward(self.single_input, state={'warmup': False})
            self.assertTrue(np.array_equal(resp, self.single_output))
        os.remove(fn_path)
        with self.subTest("Continue to use function without re-loading"):
            resp = op.forward(self.single_input, state={'warmup': False})
            self.assertTrue(np.array_equal(resp, self.single_output))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest
from io import StringIO
from unittest.mock import patch

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.op.tensorop.loss import CrossEntropy
from fastestimator.op.tensorop.model import ModelOp, UpdateOp
from fastestimator.test.unittest_util import one_layer_tf_model, sample_system_object
from fastestimator.trace.adapt import TerminateOnNaN
from fastestimator.trace.metric import Accuracy, F1Score
from fastestimator.util.data import Data


class TestTerminateOnNaN(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data_np = Data({'loss': np.NaN})
        cls.data_tf = Data({'loss': tf.constant(np.NaN)})
        cls.data_torch = Data({'loss': torch.tensor(np.NaN)})
        cls.expected_msg = "FastEstimator-TerminateOnNaN: NaN Detected in: loss"
        cls.expected_loss_keys = {"ce"}
        cls.expected_all_keys = {"ce", "accuracy", "f1_score"}

        tf_model = fe.build(model_fn=one_layer_tf_model, optimizer_fn='adam')
        cls.network = fe.Network(ops=[
            ModelOp(model=tf_model, inputs="x", outputs="y"),
            CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
            UpdateOp(model=tf_model, loss_name="ce")
        ])
        cls.traces = [
            Accuracy(true_key="y", pred_key="y_pred", output_name="accuracy"),
            F1Score(true_key="y", pred_key="y_pred", output_name="f1_score")
        ]

    def test_on_epoch_begin_loss_keys(self):
        terminate_on_nan = TerminateOnNaN(monitor_names=None)
        terminate_on_nan.system = sample_system_object()
        terminate_on_nan.system.network = self.network
        terminate_on_nan.on_epoch_begin(data={})
        self.assertEqual(terminate_on_nan.monitor_keys, self.expected_loss_keys)

    def test_on_epoch_begin_all_keys(self):
        terminate_on_nan = TerminateOnNaN(monitor_names="*")
        terminate_on_nan.system = sample_system_object()
        terminate_on_nan.system.mode = "eval"
        terminate_on_nan.system.network = self.network
        terminate_on_nan.system.traces = self.traces
        terminate_on_nan.on_epoch_begin(data={})
        self.assertEqual(terminate_on_nan.monitor_keys, self.expected_all_keys)

    def test_on_batch_end(self):
        terminate_on_nan = TerminateOnNaN(monitor_names="loss")
        terminate_on_nan.system = sample_system_object()
        terminate_on_nan.on_epoch_begin(data={})
        with self.subTest('Numpy Data'):
            with patch('sys.stdout', new=StringIO()) as fake_stdout:
                terminate_on_nan.on_batch_end(data=self.data_np)
                log = fake_stdout.getvalue().strip()
                self.assertEqual(log, self.expected_msg)

        with self.subTest('Tf Data'):
            with patch('sys.stdout', new=StringIO()) as fake_stdout:
                terminate_on_nan.on_batch_end(data=self.data_tf)
                log = fake_stdout.getvalue().strip()
                self.assertEqual(log, self.expected_msg)

        with self.subTest('Torch Data'):
            with patch('sys.stdout', new=StringIO()) as fake_stdout:
                terminate_on_nan.on_batch_end(data=self.data_torch)
                log = fake_stdout.getvalue().strip()
                self.assertEqual(log, self.expected_msg)

    def test_on_epoch_end(self):
        terminate_on_nan = TerminateOnNaN(monitor_names="loss")
        terminate_on_nan.system = sample_system_object()
        terminate_on_nan.on_epoch_begin(data={})
        with self.subTest('Numpy Data'):
            with patch('sys.stdout', new=StringIO()) as fake_stdout:
                terminate_on_nan.on_epoch_end(data=self.data_np)
                log = fake_stdout.getvalue().strip()
                self.assertEqual(log, self.expected_msg)

        with self.subTest('Tf Data'):
            with patch('sys.stdout', new=StringIO()) as fake_stdout:
                terminate_on_nan.on_epoch_end(data=self.data_tf)
                log = fake_stdout.getvalue().strip()
                self.assertEqual(log, self.expected_msg)

        with self.subTest('Torch Data'):
            with patch('sys.stdout', new=StringIO()) as fake_stdout:
                terminate_on_nan.on_epoch_end(data=self.data_torch)
                log = fake_stdout.getvalue().strip()
                self.assertEqual(log, self.expected_msg)



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf

from fastestimator.architecture.tensorflow import WideResidualNetwork


class TestWideResNet(unittest.TestCase):
    def test_wrn(self):
        data = np.ones((1, 32, 32, 3))
        input_data = tf.constant(data)
        wrn = WideResidualNetwork(input_shape=(32, 32, 3), classes=5)
        output_shape = wrn(input_data).numpy().shape
        self.assertEqual(output_shape, (1, 5))

    def test_wrn_depth(self):
        with self.assertRaises(ValueError):
            WideResidualNetwork(input_shape=(32, 32, 3), depth=27)



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tarfile
from pathlib import Path
from typing import List, Optional, Tuple

import wget

from fastestimator.dataset.numpy_dataset import NumpyDataset
from fastestimator.util.wget_util import bar_custom, callback_progress

wget.callback_progress = callback_progress


def _read_data(file_path: str) -> List[str]:
    with open(file_path) as f:
        data = [x.strip() for x in f if x.strip()]
    return data


def _create_dataset(data_path: str, translate_option: str, extension: str) -> NumpyDataset:
    source, target = translate_option.split("_to_")
    source = source.replace("_", "-")
    if extension != "train":
        source = source.split("-")[0]
    source_data = _read_data(os.path.join(data_path, source + "." + extension))
    target_data = _read_data(os.path.join(data_path, target + "." + extension))
    assert len(target_data) == len(source_data), "Sizes do not match for {} ({} mode)".format(translate_option, extension)
    dataset = NumpyDataset({"source": source_data, "target": target_data})
    return dataset


def load_data(root_dir: Optional[str] = None,
              translate_option: str = "az_to_en") -> Tuple[NumpyDataset, NumpyDataset, NumpyDataset]:
    """Load and return the neural machine translation dataset from TED talks.

    Args:
        root_dir: The path to store the downloaded data. When `path` is not provided, the data will be saved into
            `fastestimator_data` under the user's home directory.
        translate_option: Options for translation languages. Available options are: "az_to_en", "az_tr_to_en",
            "be_ru_to_en", "be_to_en", "es_to_pt", "fr_to_pt", "gl_pt_to_en", "gl_to_en", "he_to_pt", "it_to_pt",
            "pt_to_en", "ru_to_en", "ru_to_pt", and "tr_to_en".

    Returns:
        (train_data, eval_data, test_data)
    """
    # Set up path
    home = str(Path.home())
    if root_dir is None:
        root_dir = os.path.join(home, 'fastestimator_data', 'tednmt')
    else:
        root_dir = os.path.join(os.path.abspath(root_dir), 'tednmt')
    os.makedirs(root_dir, exist_ok=True)
    compressed_path = os.path.join(root_dir, 'qi18naacl-dataset.tar.gz')
    extracted_path = os.path.join(root_dir, 'datasets')
    if not os.path.exists(extracted_path):
        # Download
        if not os.path.exists(compressed_path):
            print("Downloading data to {}".format(compressed_path))
            wget.download('http://www.phontron.com/data/qi18naacl-dataset.tar.gz', compressed_path, bar=bar_custom)
        # Extract
        print("\nExtracting files ...")
        with tarfile.open(compressed_path) as f:
            f.extractall(root_dir)
    # process data
    data_path = os.path.join(extracted_path, translate_option)
    assert os.path.exists(data_path), "folder {} does not exist, please verify translation options".format(data_path)
    train_ds = _create_dataset(data_path=data_path, translate_option=translate_option, extension="train")
    eval_ds = _create_dataset(data_path=data_path, translate_option=translate_option, extension="dev")
    test_ds = _create_dataset(data_path=data_path, translate_option=translate_option, extension="test")
    return train_ds, eval_ds, test_ds



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

from fastestimator.search.golden_section import GoldenSection


class TestSearch(unittest.TestCase):
    def test_reversed_x_max_x_min(self):
        with self.assertRaises(AssertionError):
            GoldenSection(score_fn=lambda search_idx, n: (n - 3)**2, x_min=6, x_max=0, max_iter=10)

    def test_invalid_score_fn(self):
        with self.assertRaises(AssertionError):
            GoldenSection(score_fn=lambda search_idx, n, q: (n - 3)**2, x_min=0, x_max=6, max_iter=10)

    def test_integer_results(self):
        search = GoldenSection(score_fn=lambda search_idx, n: (n - 3)**2, x_min=0, x_max=6, max_iter=10, best_mode="min")
        search.fit()
        self.assertEqual(search.get_best_results()[0]["n"], 3)

    def test_float_results(self):
        search = GoldenSection(
            score_fn=lambda search_idx, n: (n - 3)**2, x_min=0, x_max=6, max_iter=10, best_mode="min", integer=False)
        search.fit()
        answer = 3.0
        result = search.get_best_results()[0]["n"]
        self.assertTrue(abs(result - answer) < 0.01)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, Set, Union

from fastestimator.backend.check_nan import check_nan
from fastestimator.schedule.schedule import get_current_items
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import traceable


@traceable()
class TerminateOnNaN(Trace):
    """End Training if a NaN value is detected.

    By default (monitor_names=None) it will monitor all loss values at the end of each batch. If one or more inputs are
    specified, it will only monitor those values. Inputs may be loss keys and/or the keys corresponding to the outputs
    of other traces (ex. accuracy).

    Args:
        monitor_names: key(s) to monitor for NaN values. If None, all loss values will be monitored. "*" will monitor
            all trace output keys and losses.
        mode: What mode(s) to execute this Trace in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Trace in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(
        self,
        monitor_names: Union[None, str, Iterable[str]] = None,
        mode: Union[None, str, Set[str]] = None,
        ds_id: Union[None, str, Iterable[str]] = None,
    ) -> None:
        super().__init__(inputs=monitor_names, mode=mode, ds_id=ds_id)
        self.monitor_keys = {}
        self.in_list = True

    def on_epoch_begin(self, data: Data) -> None:
        if not self.inputs:
            self.monitor_keys = self.system.network.get_loss_keys()
        elif "*" in self.inputs:
            self.monitor_keys = self.system.network.get_loss_keys()
            for trace in get_current_items(self.system.traces, run_modes=self.system.mode, epoch=self.system.epoch_idx):
                self.monitor_keys.update(trace.outputs)
        else:
            self.monitor_keys = self.inputs

    def on_batch_end(self, data: Data) -> None:
        for key in self.monitor_keys:
            if key in data:
                if check_nan(data[key]):
                    self.system.stop_training = True
                    print("FastEstimator-TerminateOnNaN: NaN Detected in: {}".format(key))

    def on_epoch_end(self, data: Data) -> None:
        for key in self.monitor_keys:
            if key in data:
                if check_nan(data[key]):
                    self.system.stop_training = True
                    print("FastEstimator-TerminateOnNaN: NaN Detected in: {}".format(key))



import math
import unittest

import numpy as np

import fastestimator as fe
from fastestimator.architecture.pytorch import LeNet as LeNet_torch
from fastestimator.architecture.tensorflow import LeNet as LeNet_tf
from fastestimator.backend import get_lr
from fastestimator.dataset.data import mnist
from fastestimator.op.numpyop.univariate import ExpandDims, Minmax
from fastestimator.op.tensorop.loss import CrossEntropy
from fastestimator.op.tensorop.model import ModelOp, UpdateOp
from fastestimator.schedule import ARC
from fastestimator.trace.adapt import LRScheduler


class TestLRScheduler(unittest.TestCase):

    @staticmethod
    def create_estimator_for_arc(model, use_eval, axis):
        train_data, eval_data = mnist.load_data()
        pipeline = fe.Pipeline(train_data=train_data,
                               eval_data=eval_data if use_eval else None,
                               batch_size=8,
                               ops=[ExpandDims(inputs="x", outputs="x", axis=axis), Minmax(inputs="x", outputs="x")])
        network = fe.Network(ops=[
            ModelOp(model=model, inputs="x", outputs="y_pred"),
            CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
            UpdateOp(model=model, loss_name="ce")
        ])
        estimator = fe.Estimator(pipeline=pipeline,
                                 network=network,
                                 epochs=2,
                                 traces=LRScheduler(model=model, lr_fn=ARC(1)),
                                 max_train_steps_per_epoch=10)
        return estimator

    def test_tf_model_arc_train_eval(self):
        model_tf = fe.build(model_fn=LeNet_tf, optimizer_fn="adam")
        lr_before = get_lr(model=model_tf)
        estimator = self.create_estimator_for_arc(model_tf, use_eval=True, axis=-1)
        estimator.fit()
        lr_after = get_lr(model=model_tf)
        lr_ratio = lr_after / lr_before
        increased = math.isclose(lr_ratio, 1.618, rel_tol=1e-5)
        constant = math.isclose(lr_ratio, 1.0, rel_tol=1e-5)
        decreased = math.isclose(lr_ratio, 0.618, rel_tol=1e-5)
        self.assertTrue(increased or constant or decreased)

    def test_tf_model_arc_train_only(self):
        model_tf = fe.build(model_fn=LeNet_tf, optimizer_fn="adam")
        lr_before = get_lr(model=model_tf)
        estimator = self.create_estimator_for_arc(model_tf, use_eval=False, axis=-1)
        estimator.fit()
        lr_after = get_lr(model=model_tf)
        lr_ratio = np.round(lr_after / lr_before, 3)
        increased = math.isclose(lr_ratio, 1.618, rel_tol=1e-5)
        constant = math.isclose(lr_ratio, 1.0, rel_tol=1e-5)
        decreased = math.isclose(lr_ratio, 0.618, rel_tol=1e-5)
        self.assertTrue(increased or constant or decreased)

    def test_torch_model_arc_train_eval(self):
        model_tf = fe.build(model_fn=LeNet_torch, optimizer_fn="adam")
        lr_before = get_lr(model=model_tf)
        estimator = self.create_estimator_for_arc(model_tf, use_eval=True, axis=0)
        estimator.fit()
        lr_after = get_lr(model=model_tf)
        lr_ratio = lr_after / lr_before
        increased = math.isclose(lr_ratio, 1.618, rel_tol=1e-5)
        constant = math.isclose(lr_ratio, 1.0, rel_tol=1e-5)
        decreased = math.isclose(lr_ratio, 0.618, rel_tol=1e-5)
        self.assertTrue(increased or constant or decreased)

    def test_torch_model_arc_train_only(self):
        model_tf = fe.build(model_fn=LeNet_torch, optimizer_fn="adam")
        lr_before = get_lr(model=model_tf)
        estimator = self.create_estimator_for_arc(model_tf, use_eval=False, axis=0)
        estimator.fit()
        lr_after = get_lr(model=model_tf)
        lr_ratio = lr_after / lr_before
        increased = math.isclose(lr_ratio, 1.618, rel_tol=1e-5)
        constant = math.isclose(lr_ratio, 1.0, rel_tol=1e-5)
        decreased = math.isclose(lr_ratio, 0.618, rel_tol=1e-5)
        self.assertTrue(increased or constant or decreased)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple

import tensorflow as tf

from fastestimator.dataset.numpy_dataset import NumpyDataset


def load_data(image_key: str = "x",
              label_key: str = "y",
              label_mode: str = "fine") -> Tuple[NumpyDataset, NumpyDataset]:
    """Load and return the CIFAR100 dataset.

    Please consider using the ciFAIR100 dataset instead. CIFAR100 contains duplicates between its train and test sets.

    Args:
        image_key: The key for image.
        label_key: The key for label.
        label_mode: Either "fine" for 100 classes or "coarse" for 20 classes.

    Returns:
        (train_data, eval_data)

    Raises:
        ValueError: If the label_mode is invalid.
    """
    print("\033[93m {}\033[00m".format("FastEstimator-Warn: Consider using the ciFAIR100 dataset instead."))
    if label_mode not in ['fine', 'coarse']:
        raise ValueError("label_mode must be one of either 'fine' or 'coarse'.")
    (x_train, y_train), (x_eval, y_eval) = tf.keras.datasets.cifar100.load_data(label_mode=label_mode)
    train_data = NumpyDataset({image_key: x_train, label_key: y_train})
    eval_data = NumpyDataset({image_key: x_eval, label_key: y_eval})
    return train_data, eval_data



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, Iterable, List, Optional, Tuple, TypeVar, Union

import tensorflow as tf
import torch

from fastestimator.backend.binary_crossentropy import binary_crossentropy
from fastestimator.backend.categorical_crossentropy import categorical_crossentropy
from fastestimator.backend.sparse_categorical_crossentropy import sparse_categorical_crossentropy
from fastestimator.op.tensorop.loss.loss import LossOp
from fastestimator.util.traceability_util import traceable

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


@traceable()
class CrossEntropy(LossOp):
    """Calculate Element-Wise CrossEntropy (binary, categorical or sparse categorical).

    Args:
        inputs: A tuple or list like: [<y_pred>, <y_true>].
        outputs: String key under which to store the computed loss value.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        from_logits: Whether y_pred is logits (without softmax).
        average_loss: Whether to average the element-wise loss after the Loss Op.
        form: What form of cross entropy should be performed ('binary', 'categorical', 'sparse', or None). None will
            automatically infer the correct form based on tensor shape.
        class_weights: Dictionary mapping class indices to a weight for weighting the loss function. Useful when you
            need to pay more attention to samples from an under-represented class.

    Raises:
        AssertionError: If `class_weights` or it's keys and values are of unacceptable data types.
    """
    def __init__(self,
                 inputs: Union[Tuple[str, str], List[str]],
                 outputs: str,
                 mode: Union[None, str, Iterable[str]] = "!infer",
                 ds_id: Union[None, str, Iterable[str]] = None,
                 from_logits: bool = False,
                 average_loss: bool = True,
                 form: Optional[str] = None,
                 class_weights: Optional[Dict[int, float]] = None):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id, average_loss=average_loss)
        self.from_logits = from_logits
        self.form = form
        self.cross_entropy_fn = {
            "binary": binary_crossentropy,
            "categorical": categorical_crossentropy,
            "sparse": sparse_categorical_crossentropy
        }

        if class_weights:
            assert isinstance(class_weights, dict), \
                "class_weights should be a dictionary or have None value, got {}".format(type(class_weights))
            assert all(isinstance(key, int) for key in class_weights.keys()), \
                "Please ensure that the keys of the class_weight dictionary are of type: int"
            assert all(isinstance(value, float) for value in class_weights.values()), \
                "Please ensure that the values of the class_weight dictionary are of type: float"

        self.class_weights = class_weights
        self.class_dict = None

    def build(self, framework: str, device: Optional[torch.device] = None) -> None:
        if self.class_weights:
            if framework == 'tf':
                keys_tensor = tf.constant(list(self.class_weights.keys()))
                vals_tensor = tf.constant(list(self.class_weights.values()))
                self.class_dict = tf.lookup.StaticHashTable(
                    tf.lookup.KeyValueTensorInitializer(keys_tensor, vals_tensor), default_value=1.0)
            elif framework == 'torch':
                self.class_dict = self.class_weights
            else:
                raise ValueError("unrecognized framework: {}".format(framework))

    def forward(self, data: List[Tensor], state: Dict[str, Any]) -> Tensor:
        y_pred, y_true = data
        form = self.form
        if form is None:
            if len(y_pred.shape) == 2 and y_pred.shape[-1] > 1:
                if len(y_true.shape) == 2 and y_true.shape[-1] > 1:
                    form = "categorical"
                else:
                    form = "sparse"
            else:
                form = "binary"

        loss = self.cross_entropy_fn[form](y_pred,
                                           y_true,
                                           from_logits=self.from_logits,
                                           average_loss=self.average_loss,
                                           class_weights=self.class_dict)
        return loss



#  Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
import os
import shutil
import tempfile
import unittest
from contextlib import closing
from datetime import datetime

from fastestimator.summary.history import connect, delete, update_settings

DEFAULT_KEEP = 500
DEFAULT_LOG_KEEP = 500


class TestConnect(unittest.TestCase):
    def test_make_schema(self):
        db = connect(":memory:")
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT name FROM sqlite_master WHERE type = 'table'")
            results = cursor.fetchall()
        results = {result['name'] for result in results}
        expected = {"datasets", "errors", "features", "history", "pipeline", "network", "postprocess", "traces",
                    "errors", "logs", "settings"}
        self.assertSetEqual(results, expected)
        db.close()

    def test_fk_delete(self):
        db = connect(":memory:")
        db.execute("INSERT INTO history (pk) VALUES (0)")
        db.execute("INSERT INTO network (fk) VALUES (0)")
        db.commit()
        with closing(db.cursor()) as cursor:
            with self.subTest("entry should be inserted correctly"):
                cursor.execute("SELECT count(*) AS count FROM network")
                results = cursor.fetchall()
                self.assertEqual(results[0]['count'], 1)
            with self.subTest("entry should cascade delete"):
                db.execute("DELETE FROM history WHERE pk = 0")
                cursor.execute("SELECT count(*) AS count FROM network")
                results = cursor.fetchall()
                self.assertEqual(results[0]['count'], 0)
        db.close()

    def test_fk_update(self):
        db = connect(":memory:")
        db.execute("INSERT INTO history (pk) VALUES (0)")
        db.execute("INSERT INTO network (fk) VALUES (0)")
        db.commit()
        with closing(db.cursor()) as cursor:
            with self.subTest("entry should be inserted correctly"):
                cursor.execute("SELECT * FROM network")
                results = cursor.fetchall()
                self.assertEqual(results[0]['fk'], 0)
            with self.subTest("entry should cascade update"):
                db.execute("UPDATE history SET pk = 10 WHERE pk = 0")
                cursor.execute("SELECT * FROM network")
                results = cursor.fetchall()
                self.assertEqual(results[0]['fk'], 10)
        db.close()

    def test_initial_settings(self):
        db = connect(":memory:")
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM settings")
            results = cursor.fetchall()
        with self.subTest("should be exactly one setting row"):
            self.assertEqual(len(results), 1)
        results = results[0]
        with self.subTest("settings should have pk = 0"):
            self.assertEqual(results['pk'], 0)
        with self.subTest("schema version should be 1"):
            self.assertEqual(results['schema_version'], 1)
        with self.subTest(f"n_keep should be {DEFAULT_KEEP}"):
            self.assertEqual(results['n_keep'], DEFAULT_KEEP)
        with self.subTest(f"n_keep_logs should be {DEFAULT_LOG_KEEP}"):
            self.assertEqual(results['n_keep_logs'], DEFAULT_LOG_KEEP)


class TestDelete(unittest.TestCase):
    def setUp(self):
        self.db_dir = tempfile.mkdtemp()
        self.db_path = os.path.join(self.db_dir, 'tmp.db')

    def tearDown(self):
        shutil.rmtree(self.db_dir)

    def test_under_threshold(self):
        db = connect(self.db_path)
        for i in range(10):
            db.execute("INSERT INTO history (pk, train_start) VALUES (?, ?)", [i, datetime.now()])
        db.commit()
        with closing(db.cursor()) as cursor:
            with self.subTest("entries should be inserted correctly"):
                cursor.execute("SELECT count(*) AS count FROM history")
                results = cursor.fetchall()
                self.assertEqual(results[0]['count'], 10)
            delete(n_keep=10, db_path=self.db_path)
            with self.subTest("no entries should be deleted"):
                cursor.execute("SELECT count(*) AS count FROM history")
                results = cursor.fetchall()
                self.assertEqual(results[0]['count'], 10)
        db.close()

    def test_over_threshold(self):
        db = connect(self.db_path)
        for i in range(10):
            db.execute("INSERT INTO history (pk, train_start) VALUES (?, ?)", [i, datetime.now()])
        db.commit()
        with closing(db.cursor()) as cursor:
            with self.subTest("entries should be inserted correctly"):
                cursor.execute("SELECT count(*) AS count FROM history")
                results = cursor.fetchall()
                self.assertEqual(results[0]['count'], 10)
            delete(n_keep=9, db_path=self.db_path)
            with self.subTest("one entry should be deleted"):
                cursor.execute("SELECT count(*) AS count FROM history")
                results = cursor.fetchall()
                self.assertEqual(results[0]['count'], 9)
            with self.subTest("oldest entry should be deleted"):
                cursor.execute("SELECT pk FROM history")
                results = cursor.fetchall()
                results = {result['pk'] for result in results}
                expected = {i for i in range(1, 10)}
                self.assertSetEqual(results, expected)
        db.close()


class TestUpdateSettings(unittest.TestCase):
    def setUp(self):
        self.db_dir = tempfile.mkdtemp()
        self.db_path = os.path.join(self.db_dir, 'tmp.db')

    def tearDown(self):
        shutil.rmtree(self.db_dir)

    def test_decrease_n_keep(self):
        update_settings(n_keep=42, db_path=self.db_path)
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM settings WHERE pk = 0")
            results = cursor.fetchall()[0]
        with self.subTest("n_keep updated"):
            self.assertEqual(results['n_keep'], 42)
        with self.subTest("n_keep_logs also reduced"):
            self.assertEqual(results['n_keep_logs'], 42)

    def test_increase_n_keep(self):
        update_settings(n_keep=2000, db_path=self.db_path)
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM settings WHERE pk = 0")
            results = cursor.fetchall()[0]
        with self.subTest("n_keep updated"):
            self.assertEqual(results['n_keep'], 2000)
        with self.subTest("n_keep_logs not updated"):
            self.assertEqual(results['n_keep_logs'], DEFAULT_LOG_KEEP)

    def test_decrease_logs(self):
        update_settings(n_keep_logs=42, db_path=self.db_path)
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM settings WHERE pk = 0")
            results = cursor.fetchall()[0]
        with self.subTest("n_keep not updated"):
            self.assertEqual(results['n_keep'], DEFAULT_KEEP)
        with self.subTest("n_keep_logs updated"):
            self.assertEqual(results['n_keep_logs'], 42)

    def test_increase_logs(self):
        update_settings(n_keep_logs=1500, db_path=self.db_path)
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM settings WHERE pk = 0")
            results = cursor.fetchall()[0]
        with self.subTest("n_keep not updated"):
            self.assertEqual(results['n_keep'], DEFAULT_KEEP)
        with self.subTest("n_keep_logs not updated (n_logs can't exceed n_keep)"):
            self.assertEqual(results['n_keep_logs'], DEFAULT_LOG_KEEP)

    def test_joint_update_raise_kgl(self):
        update_settings(n_keep=1200, n_keep_logs=1100, db_path=self.db_path)
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM settings WHERE pk = 0")
            results = cursor.fetchall()[0]
        with self.subTest("n_keep updated"):
            self.assertEqual(results['n_keep'], 1200)
        with self.subTest("n_keep_logs updated"):
            self.assertEqual(results['n_keep_logs'], 1100)

    def test_joint_update_raise_lgk(self):
        update_settings(n_keep=1100, n_keep_logs=1200, db_path=self.db_path)
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM settings WHERE pk = 0")
            results = cursor.fetchall()[0]
        with self.subTest("n_keep updated"):
            self.assertEqual(results['n_keep'], 1100)
        with self.subTest("n_keep_logs updated"):
            self.assertEqual(results['n_keep_logs'], 1100)

    def test_joint_update_lower_lgk(self):
        update_settings(n_keep=300, n_keep_logs=400, db_path=self.db_path)
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM settings WHERE pk = 0")
            results = cursor.fetchall()[0]
        with self.subTest("n_keep updated"):
            self.assertEqual(results['n_keep'], 300)
        with self.subTest("n_keep_logs updated"):
            self.assertEqual(results['n_keep_logs'], 300)

    def test_joint_update_lower_kgl(self):
        update_settings(n_keep=400, n_keep_logs=300, db_path=self.db_path)
        db = connect(self.db_path)
        with closing(db.cursor()) as cursor:
            cursor.execute("SELECT * FROM settings WHERE pk = 0")
            results = cursor.fetchall()[0]
        with self.subTest("n_keep updated"):
            self.assertEqual(results['n_keep'], 400)
        with self.subTest("n_keep_logs updated"):
            self.assertEqual(results['n_keep_logs'], 300)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import sys
from typing import Callable

import wget


def bar_custom(current: float, total: float, width: int = 80) -> str:
    """Return progress bar string for given values in one of three styles depending on available width.

    This function was modified from wget source code at https://bitbucket.org/techtonik/python-wget/src/default/.

    The bar will be one of the following formats depending on available width:
        [..  ] downloaded / total
        downloaded / total
        [.. ]

    If total width is unknown or <= 0, the bar will show a bytes counter using two adaptive styles:
        %s / unknown
        %s

    If there is not enough space on the screen, do not display anything. The returned string doesn't include control
    characters like \r used to place cursor at the beginning of the line to erase previous content.

    This function leaves one free character at the end of the string to avoid automatic linefeed on Windows.

    ```python
    wget.download('http://url.com', '/save/dir', bar=fe.util.bar_custom)
    ```

    Args:
        current: The current amount of progress.
        total: The total amount of progress required by the task.
        width: The available width.

    Returns:
        A formatted string to display the current progress.
    """
    # process special case when total size is unknown and return immediately
    if not total or total < 0:
        msg = "{} / unknown".format(current)
        if len(msg) < width:  # leaves one character to avoid linefeed
            return msg
        if len("{}".format(current)) < width:
            return "{}".format(current)

    # --- adaptive layout algorithm ---
    #
    # [x] describe the format of the progress bar
    # [x] describe min width for each data field
    # [x] set priorities for each element
    # [x] select elements to be shown
    #   [x] choose top priority element min_width < avail_width
    #   [x] lessen avail_width by value if min_width
    #   [x] exclude element from priority list and repeat

    #  10% [.. ]  10/100
    # pppp bbbbb sssssss

    min_width = {
        'percent': 4,  # 100%
        'bar': 3,  # [.]
        'size': len("{}".format(total)) * 2 + 3,  # 'xxxx / yyyy'
    }
    priority = ['percent', 'bar', 'size']

    # select elements to show
    selected = []
    avail = width
    for field in priority:
        if min_width[field] < avail:
            selected.append(field)
            avail -= min_width[field] + 1  # +1 is for separator or for reserved space at
            # the end of line to avoid linefeed on Windows

    # render
    output = ''
    for field in selected:

        if field == 'percent':
            # fixed size width for percentage
            output += "{}%".format(100 * current // total).rjust(min_width['percent'])
        elif field == 'bar':  # [. ]
            # bar takes its min width + all available space
            output += wget.bar_thermometer(current, total, min_width['bar'] + avail)
        elif field == 'size':
            # size field has a constant width (min == max)
            output += "{:.2f} / {:.2f} MB".format(current / 1e6, total / 1e6).rjust(min_width['size'])

        selected = selected[1:]
        if selected:
            output += ' '  # add field separator

    return output


def callback_progress(blocks: int, block_size: int, total_size: int, bar_function: Callable[[int, int, int],
                                                                                            str]) -> None:
    """Callback function for urlretrieve that is called when a connection is created and then once for each block.

    Draws adaptive progress bar in terminal/console.

    Use sys.stdout.write() instead of "print", because it allows one more symbols at the line end without triggering a
    linefeed on Windows.

    ```python
    import wget
    wget.callback_progress = fe.util.callback_progress
    wget.download('http://url.com', '/save/dir', bar=fe.util.bar_custom)
    ```

    Args:
        blocks: number of blocks transferred so far.
        block_size: in bytes.
        total_size: in bytes, can be -1 if server doesn't return it.
        bar_function: another callback function to visualize progress.
    """
    width = min(100, wget.get_console_width())
    if width == 0:  # sys.stdout.fileno() in get_console_width() is not supported in jupyter notebook
        width = 80

    current_size = min(blocks * block_size, total_size)

    progress = bar_function(current_size, total_size, width)
    if progress:
        sys.stdout.write("\r{}".format(progress))
        if current_size >= total_size:
            sys.stdout.write("\n")



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import random
import tarfile
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import pandas as pd
import wget

from fastestimator.dataset.csv_dataset import CSVDataset
from fastestimator.util.wget_util import bar_custom, callback_progress

wget.callback_progress = callback_progress


def _create_csv(images: List[str], label_dict: Dict[str, int], csv_path: str) -> None:
    """A helper function to create and save csv files.

    Args:
        images: List of image id's.
        label_dict: Mapping of class name to label id.
        csv_path: Path to save the csv file.
    """

    df = pd.DataFrame(images, columns=["image"])
    df["label"] = df["image"].apply(lambda x: label_dict[x.split("/")[0]])
    df["image"] = "food-101/images/" + df["image"] + ".jpg"
    df.to_csv(csv_path, index=False)
    return None


def load_data(root_dir: Optional[str] = None) -> Tuple[CSVDataset, CSVDataset]:
    """Load and return the Food-101 dataset.

    Food-101 dataset is a collection of images from 101 food categories.
    Sourced from http://data.vision.ee.ethz.ch/cvl/food-101.tar.gz

    Args:
        root_dir: The path to store the downloaded data. When `path` is not provided, the data will be saved into
            `fastestimator_data` under the user's home directory.

    Returns:
        (train_data, test_data)
    """
    home = str(Path.home())

    if root_dir is None:
        root_dir = os.path.join(home, 'fastestimator_data', 'Food_101')
    else:
        root_dir = os.path.join(os.path.abspath(root_dir), 'Food_101')
    os.makedirs(root_dir, exist_ok=True)

    image_compressed_path = os.path.join(root_dir, 'food-101.tar.gz')
    image_extracted_path = os.path.join(root_dir, 'food-101')

    train_csv_path = os.path.join(root_dir, 'train.csv')
    test_csv_path = os.path.join(root_dir, 'test.csv')

    if not os.path.exists(image_extracted_path):
        # download
        if not os.path.exists(image_compressed_path):
            print("Downloading data to {}".format(root_dir))
            wget.download('http://data.vision.ee.ethz.ch/cvl/food-101.tar.gz', root_dir, bar=bar_custom)

        # extract
        print("\nExtracting files ...")
        with tarfile.open(image_compressed_path) as img_tar:
            img_tar.extractall(root_dir)

    labels = open(os.path.join(root_dir, "food-101/meta/classes.txt"), "r").read().split()
    label_dict = {labels[i]: i for i in range(len(labels))}

    if not os.path.exists(train_csv_path):
        train_images = open(os.path.join(root_dir, "food-101/meta/train.txt"), "r").read().split()
        random.shuffle(train_images)
        _create_csv(train_images, label_dict, train_csv_path)
    if not os.path.exists(test_csv_path):
        test_images = open(os.path.join(root_dir, "food-101/meta/test.txt"), "r").read().split()
        random.shuffle(test_images)
        _create_csv(test_images, label_dict, test_csv_path)

    return CSVDataset(train_csv_path), CSVDataset(test_csv_path)



import tempfile
import unittest

import fastestimator as fe
from fastestimator.op.numpyop import NumpyOp
from fastestimator.test.unittest_util import sample_system_object, sample_system_object_torch


class TestNumpyOp(NumpyOp):
    def __init__(self, inputs, outputs, mode, var):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.var = var


class TestOneOf(unittest.TestCase):
    def test_save_and_load_state_tf(self):
        def instantiate_system():
            system = sample_system_object()
            system.pipeline.ops = [
                fe.op.numpyop.meta.OneOf(
                    TestNumpyOp(inputs="x", outputs="x", mode="train", var=1),
                    TestNumpyOp(inputs="x", outputs="x", mode="train", var=1),
                )
            ]
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.pipeline.ops[0].ops[0].var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.pipeline.ops[0].ops[0].var

        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_torch(self):
        def instantiate_system():
            system = sample_system_object_torch()
            system.pipeline.ops = [
                fe.op.numpyop.meta.OneOf(
                    TestNumpyOp(inputs="x", outputs="x", mode="train", var=1),
                    TestNumpyOp(inputs="x", outputs="x", mode="train", var=1),
                )
            ]
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.pipeline.ops[0].ops[0].var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.pipeline.ops[0].ops[0].var

        self.assertEqual(loaded_var, new_var)



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Optional, Tuple

import tensorflow as tf
from tensorflow.keras import Model, layers


# noinspection PyPep8Naming
def WideResidualNetwork(input_shape: Tuple[int, int, int],
                        depth: int = 28,
                        widen_factor: int = 10,
                        dropout: float = 0.0,
                        classes: int = 10,
                        activation: Optional[str] = 'softmax') -> tf.keras.Model:
    """Creates a Wide Residual Network with specified parameters.

    Args:
        input_shape: The size of the input tensor (height, width, channels).
        depth: Depth of the network. Compute N = (n - 4) / 6.
               For a depth of 16, n = 16, N = (16 - 4) / 6 = 2
               For a depth of 28, n = 28, N = (28 - 4) / 6 = 4
               For a depth of 40, n = 40, N = (40 - 4) / 6 = 6
        widen_factor: Width of the network.
        dropout: Adds dropout if value is greater than 0.0.
        classes: The number of outputs the model should generate.
        activation: activation function for last dense layer.

    Returns:
        A Keras Model.

    Raises:
        ValueError: If (depth - 4) is not divisible by 6.
    """
    if (depth - 4) % 6 != 0:
        raise ValueError('Depth of the network must be such that (depth - 4)' 'should be divisible by 6.')

    img_input = layers.Input(shape=input_shape)
    inputs = img_input

    x = __create_wide_residual_network(classes, img_input, depth, widen_factor, dropout, activation)
    # Create model.
    model = Model(inputs, x)
    return model


def __conv1_block(inputs: tf.Tensor, n_filters: int) -> tf.Tensor:
    """Conv block of the network.

    Args:
        inputs: input tensor.
        n_filters: How many filters for the convolution layer.

    Returns:
        Output tensor of the conv block.
    """
    x = layers.Conv2D(n_filters, (3, 3), padding='same', kernel_initializer='he_normal', use_bias=False)(inputs)
    x = layers.BatchNormalization(momentum=0.9)(x)
    x = layers.Activation('relu')(x)
    return x


def __basic_block(inputs: tf.Tensor, n_filters: int = 16, dropout: float = 0.0) -> tf.Tensor:
    """Basic block of the network.

    Args:
        inputs: input tensor.
        n_filters: How many filters for the convolution layer.
        dropout: Adds dropout if value is greater than 0.0.

    Returns:
        Output tensor of the basic block.
    """
    init = inputs

    # Check if input number of filters is same as 16 * k, else create
    # convolution2d for this input
    if init.shape[-1] != n_filters:
        init = layers.Conv2D(n_filters, (1, 1),
                             activation='linear',
                             padding='same',
                             kernel_initializer='he_normal',
                             use_bias=False)(init)

    x = layers.Conv2D(n_filters, (3, 3), padding='same', kernel_initializer='he_normal', use_bias=False)(inputs)
    x = layers.BatchNormalization(momentum=0.9)(x)
    x = layers.Activation('relu')(x)

    if dropout > 0.0:
        x = layers.Dropout(dropout)(x)

    x = layers.Conv2D(n_filters, (3, 3), padding='same', kernel_initializer='he_normal', use_bias=False)(x)
    x = layers.BatchNormalization(momentum=0.9)(x)
    x = layers.Activation('relu')(x)

    m = init + x
    return m


def __create_wide_residual_network(classes: int,
                                   img_input: tf.Tensor,
                                   depth: int = 28,
                                   widen_factor: int = 10,
                                   dropout: float = 0.0,
                                   activation: Optional[str] = 'softmax') -> tf.Tensor:
    """Generates the output tensor of the Wide Residual Network.

    Args:
        classes: Number of output classes.
        img_input: Input tensor or layer.
        depth: Depth of the network. Compute N = (n - 4) / 6.
               For a depth of 16, n = 16, N = (16 - 4) / 6 = 2
               For a depth of 28, n = 28, N = (28 - 4) / 6 = 4
               For a depth of 40, n = 40, N = (40 - 4) / 6 = 6
        widen_factor: Width of the network.
        dropout: Adds dropout if value is greater than 0.0.
        activation: activation function for the last dense layer.

    Returns:
        Output tensor of the network.
    """
    width = [16, 16 * widen_factor, 32 * widen_factor, 64 * widen_factor]
    N = (depth - 4) // 6
    x = __conv1_block(img_input, width[0])

    for _ in range(N):
        x = __basic_block(x, width[1], dropout)

    x = layers.MaxPooling2D((2, 2))(x)

    for _ in range(N):
        x = __basic_block(x, width[2], dropout)

    x = layers.MaxPooling2D((2, 2))(x)

    for _ in range(N):
        x = __basic_block(x, width[3], dropout)

    x = layers.GlobalAveragePooling2D()(x)
    x = layers.Dense(classes, activation=activation)(x)
    return x



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple, Union

import torch
import torch.nn as nn


class Cropping2D(nn.Module):
    """A layer for cropping along height and width dimensions.

    This class is intentionally not @traceable (models and layers are handled by a different process).

    ```python
    x = torch.tensor(list(range(100))).view((1,1,10,10))
    m = fe.layers.pytorch.Cropping2D(3)
    y = m.forward(x)  # [[[[33, 34, 35, 36], [43, 44, 45, 46], [53, 54, 55, 56], [63, 64, 65, 66]]]]
    m = fe.layers.pytorch.Cropping2D((3, 4))
    y = m.forward(x)  # [[[[34, 35], [44, 45], [54, 55], [64, 65]]]]
    m = fe.layers.pytorch.Cropping2D(((1, 4), 4))
    y = m.forward(x)  # [[[[14, 15], [24, 25], [34, 35], [44, 45], [54, 55]]]]
    ```

    Args:
        cropping: Height and width cropping parameters. If a single int 'n' is specified, then the width and height of
            the input will both be reduced by '2n', with 'n' coming off of each side of the input. If a tuple ('h', 'w')
            is provided, then the height and width of the input will be reduced by '2h' and '2w' respectively, with 'h'
            and 'w' coming off of each side of the input. If a tuple like (('h1', 'h2'), ('w1', 'w2')) is provided, then
            'h1' will be removed from the top, 'h2' from the bottom, 'w1' from the left, and 'w2' from the right
            (assuming the top left corner as the 0,0 origin).

    Raises:
        ValueError: If `cropping` has an unacceptable data type.
    """
    def __init__(self, cropping: Union[int, Tuple[Union[int, Tuple[int, int]], Union[int, Tuple[int,
                                                                                                int]]]] = 0) -> None:
        super().__init__()

        if isinstance(cropping, int):
            self.cropping = ((cropping, cropping), (cropping, cropping))
        elif hasattr(cropping, '__len__'):
            if len(cropping) != 2:
                raise ValueError(f"'cropping' should have two elements, but found {len(cropping)}")
            if isinstance(cropping[0], int):
                height_cropping = (cropping[0], cropping[0])
            elif hasattr(cropping[0], '__len__') and len(cropping[0]) == 2:
                height_cropping = (cropping[0][0], cropping[0][1])
            else:
                raise ValueError(f"'cropping' height should be an int or tuple of ints, but found {cropping[0]}")
            if isinstance(cropping[1], int):
                width_cropping = (cropping[1], cropping[1])
            elif hasattr(cropping[1], '__len__') and len(cropping[1]) == 2:
                width_cropping = (cropping[1][0], cropping[1][1])
            else:
                raise ValueError(f"'cropping' width should be an int or tuple of ints, but found {cropping[1]}")
            self.cropping = (height_cropping, width_cropping)
        else:
            raise ValueError(
                "cropping` should be either an int, a tuple of 2 ints or a tuple of two tuple of 2 ints. Found: " +
                str(cropping))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x[:, :, self.cropping[0][0]:-self.cropping[0][1], self.cropping[1][0]:-self.cropping[1][1]]



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, List, Optional, Tuple, TypeVar, Union

from albumentations import BboxParams, KeypointParams
from albumentations.imgaug.transforms import IAAAffine

from fastestimator.op.numpyop.multivariate.multivariate import MultiVariateAlbumentation
from fastestimator.util.traceability_util import traceable

Number = TypeVar('Number', int, float)


@traceable()
class Affine(MultiVariateAlbumentation):
    """Perform affine transformations on an image.

    Args:
        rotate: How much to rotate an image (in degrees). If a single value is given then images will be rotated by
                a value sampled from the range [-n, n]. If a tuple (a, b) is given then each image will be rotated
                by a value sampled from the range [a, b].
        scale: How much to scale an image (in percentage). If a single value is given then all images will be scaled
                by a value drawn from the range [1.0, n]. If a tuple (a,b) is given then each image will be scaled
                based on a value drawn from the range [a,b].
        shear: How much to shear an image (in degrees). If a single value is given then all images will be sheared
                on X and Y by two values sampled from the range [-n, n]. If a tuple (a, b) is given then images will
                be sheared on X and Y by two values randomly sampled from the range [a, b].
        translate: How much to translate an image. If a single value is given then the translation extent will be
                sampled from the range [0,n]. If a tuple (a,b) is given then the extent will be sampled from
                the range [a,b]. If integers are given then the translation will be in pixels. If a float then
                it will be as a fraction of the image size.
        border_handling: What to do in order to fill newly created pixels. Options are 'constant', 'edge',
                'symmetric', 'reflect', and 'wrap'. If a list is given, then the method will be randomly
                selected from the options in the list.
        fill_value: What pixel value to insert when border_handling is 'constant'.
        interpolation: What interpolation method to use. Options (from fast to slow) are 'nearest_neighbor',
                'bilinear', 'bicubic', 'biquartic', and 'biquintic'.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        image_in: The key of an image to be modified.
        mask_in: The key of a mask to be modified (with the same random factors as the image).
        masks_in: The key of masks to be modified (with the same random factors as the image).
        bbox_in: The key of a bounding box(es) to be modified (with the same random factors as the image).
        keypoints_in: The key of keypoints to be modified (with the same random factors as the image).
        image_out: The key to write the modified image (defaults to `image_in` if None).
        mask_out: The key to write the modified mask (defaults to `mask_in` if None).
        masks_out: The key to write the modified masks (defaults to `masks_in` if None).
        bbox_out: The key to write the modified bounding box(es) (defaults to `bbox_in` if None).
        keypoints_out: The key to write the modified keypoints (defaults to `keypoints_in` if None).
        bbox_params: Parameters defining the type of bounding box ('coco', 'pascal_voc', 'albumentations' or 'yolo').
        keypoint_params: Parameters defining the type of keypoints ('xy', 'yx', 'xya', 'xys', 'xyas', 'xysa').

    Image types:
        uint8, float32
    """
    def __init__(self,
                 rotate: Union[Number, Tuple[Number, Number]] = 0,
                 scale: Union[float, Tuple[float, float]] = 1.0,
                 shear: Union[Number, Tuple[Number, Number]] = 0,
                 translate: Union[Number, Tuple[Number, Number]] = 0,
                 border_handling: Union[str, List[str]] = "reflect",
                 fill_value: Number = 0,
                 interpolation: str = "bilinear",
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 image_in: Optional[str] = None,
                 mask_in: Optional[str] = None,
                 masks_in: Optional[str] = None,
                 bbox_in: Optional[str] = None,
                 keypoints_in: Optional[str] = None,
                 image_out: Optional[str] = None,
                 mask_out: Optional[str] = None,
                 masks_out: Optional[str] = None,
                 bbox_out: Optional[str] = None,
                 keypoints_out: Optional[str] = None,
                 bbox_params: Union[BboxParams, str, None] = None,
                 keypoint_params: Union[KeypointParams, str, None] = None):
        order = {'nearest_neighbor': 0, 'bilinear': 1, 'bicubic': 3, 'biquartic': 4, 'biquintic': 5}[interpolation]
        if isinstance(translate, int) or (isinstance(translate, Tuple) and isinstance(translate[0], int)):
            func = IAAAffine(rotate=rotate,
                             scale=scale,
                             shear=shear,
                             translate_px=translate,
                             order=order,
                             cval=fill_value,
                             mode=border_handling,
                             always_apply=True)
        else:
            func = IAAAffine(rotate=rotate,
                             scale=scale,
                             shear=shear,
                             translate_percent=translate,
                             order=order,
                             cval=fill_value,
                             mode=border_handling,
                             always_apply=True)
        super().__init__(func,
                         image_in=image_in,
                         mask_in=mask_in,
                         masks_in=masks_in,
                         bbox_in=bbox_in,
                         keypoints_in=keypoints_in,
                         image_out=image_out,
                         mask_out=mask_out,
                         masks_out=masks_out,
                         bbox_out=bbox_out,
                         keypoints_out=keypoints_out,
                         bbox_params=bbox_params,
                         keypoint_params=keypoint_params,
                         mode=mode,
                         ds_id=ds_id)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tempfile
import unittest

import matplotlib.pyplot as plt
import numpy as np

from fastestimator.test.unittest_util import sample_system_object
from fastestimator.trace.io import ImageSaver
from fastestimator.util.data import Data
from fastestimator.util.img_data import ImgData


class TestImageSaver(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.image_dir = tempfile.gettempdir()
        cls.image_path = os.path.join(cls.image_dir, 'img_train_epoch_0_elem_0.png')
        cls.img_data_path = os.path.join(cls.image_dir, 'img_data_train_epoch_0.png')
        cls.input_img = 0.5 * np.ones((1, 32, 32, 3))
        cls.mask = np.zeros_like(cls.input_img)
        cls.mask[0, 10:20, 10:30, :] = [1, 0, 0]
        bbox = np.array([[[3, 7, 10, 6, 'box1'], [20, 20, 8, 8, 'box2']]] * 1)
        d = ImgData(y=np.ones((1, )), x=[cls.input_img, cls.mask, bbox])
        cls.data = Data({'img': cls.input_img, 'img_data': d})

    def test_on_epoch_end(self):
        image_saver = ImageSaver(inputs='img', save_dir=self.image_dir)
        image_saver.system = sample_system_object()
        image_saver.on_epoch_end(data=self.data)
        with self.subTest('Check if image is saved'):
            self.assertTrue(os.path.exists(self.image_path))
        with self.subTest('Check image is valid or not'):
            im = plt.imread(self.image_path)
            self.assertFalse(np.any(im[:, 0] == np.nan))

    def test_on_epoch_end_img_data(self):
        if os.path.exists(self.img_data_path):
            os.remove(self.img_data_path)
        image_saver = ImageSaver(inputs='img_data', save_dir=self.image_dir)
        image_saver.system = sample_system_object()
        image_saver.on_epoch_end(data=self.data)
        with self.subTest('Check if image is saved'):
            self.assertTrue(os.path.exists(self.img_data_path))
        with self.subTest('Check image is valid or not'):
            im = plt.imread(self.img_data_path)
            self.assertFalse(np.any(im[:, 0] == np.nan))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

from fastestimator.op.op import Op, get_inputs_by_op, write_outputs_by_op


class TestOp(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.testop = Op(inputs="x", outputs="y", mode="train")

    def test_op_initialization(self):
        self.assertEqual(self.testop.inputs, ["x"])
        self.assertEqual(self.testop.outputs, ["y"])
        self.assertEqual(self.testop.mode, {"train"})


class TestGetInputsByOp(unittest.TestCase):
    def test_single_key_single_data(self):
        data = get_inputs_by_op(op=Op(inputs="x"), store={"x": 1})
        self.assertEqual(data, 1)

    def test_single_key_multi_data(self):
        data = get_inputs_by_op(op=Op(inputs="x"), store={"x": [1, 2]})
        self.assertEqual(data, [1, 2])

    def test_multi_key_multi_data(self):
        x, y = get_inputs_by_op(op=Op(inputs=["x", "y"]), store={"x": 1, "y": [1, 2]})
        self.assertEqual(x, 1)
        self.assertEqual(y, [1, 2])


class TestWriteOutputsByOp(unittest.TestCase):
    def test_single_key_single_data(self):
        batch = {}
        write_outputs_by_op(op=Op(outputs="x"), store=batch, outputs=1)
        self.assertEqual(batch, {"x": 1})

    def test_single_key_multi_data(self):
        batch = {}
        write_outputs_by_op(op=Op(outputs="x"), store=batch, outputs=[1, 2])
        self.assertEqual(batch, {"x": [1, 2]})

    def test_multi_key_multi_data(self):
        batch = {}
        write_outputs_by_op(op=Op(outputs=["x", "y"]), store=batch, outputs=[1, [1, 2]])
        self.assertEqual(batch, {"x": 1, "y": [1, 2]})



#  Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
import tempfile
import unittest
from collections import defaultdict

import tensorflow as tf

from fastestimator.test.unittest_util import sample_system_object
from fastestimator.trace.xai import InstanceTracker
from fastestimator.util import Data


class TestInstanceTracker(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Epoch 1 Data
        batch1_metrics = tf.constant([0.9, 0.8, 0.1, 0.4, 0.9, 0.6, 0.1, 0.6, 0.8, 0.3])
        batch1_idx = tf.constant([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        batch2_metrics = tf.constant([0.3, 0.1, 0.0, 0.8, 0.5, 0.5, 0.6, 0.2, 0.5, 1.0])
        batch2_idx = tf.constant([10, 11, 12, 13, 14, 15, 16, 17, 18, 19])
        batch3_metrics = tf.constant([0.3, 0.4, 0.7, 0.9, 0.3, 0.0, 0.3, 0.7, 0.8, 0.6])
        batch3_idx = tf.constant([20, 21, 22, 23, 24, 25, 26, 27, 28, 29])
        # Epoch 2 Data
        batch4_metrics = tf.constant([0.8, 0.9, 0.6, 0.1, 0.7, 0.3, 0.9, 0.9, 0.4, 0.6])
        batch4_idx = tf.constant([21, 2, 18, 3, 15, 22, 12, 27, 23, 9])
        batch5_metrics = tf.constant([0.4, 0.4, 0.9, 0.1, 0.4, 0.9, 0.0, 0.8, 1.0, 0.1])
        batch5_idx = tf.constant([20, 5, 28, 8, 6, 25, 11, 13, 7, 16])
        batch6_metrics = tf.constant([0.3, 0.9, 0.9, 0.5, 0.9, 0.8, 0.6, 0.1, 0.1, 0.2])
        batch6_idx = tf.constant([1, 19, 24, 10, 0, 29, 17, 14, 4, 26])

        cls.training_data = [[
            Data({'ce': batch1_metrics, 'idx': batch1_idx}),
            Data({'ce': batch2_metrics, 'idx': batch2_idx}),
            Data({'ce': batch3_metrics, 'idx': batch3_idx})
        ], [Data({'ce': batch4_metrics, 'idx': batch4_idx}),
            Data({'ce': batch5_metrics, 'idx': batch5_idx}),
            Data({'ce': batch6_metrics, 'idx': batch6_idx})
            ]]

    def _simulate_training(self, trace: InstanceTracker, data: Data):
        trace.on_begin(Data())
        system = trace.system
        global_step = 0
        for epoch_idx, epoch in enumerate(self.training_data, start=1):
            system.epoch_idx = epoch_idx
            trace.on_epoch_begin(Data())
            for batch_idx, batch in enumerate(epoch, start=1):
                system.batch_idx = batch_idx
                global_step += 1
                system.global_step = global_step
                trace.on_batch_begin(Data())
                trace.on_batch_end(batch)
            trace.on_epoch_end(Data())
        trace.on_end(data)

    def test_basic_happy_path(self):
        instance_tracker = InstanceTracker(index='idx', metric='ce', outputs='out', mode='train')
        system = sample_system_object()
        instance_tracker.system = system
        response = Data()
        self._simulate_training(instance_tracker, response)

        # Check responses
        with self.subTest('Check that a response was written'):
            self.assertIn('out', response)
        with self.subTest('Check that data was written to the system'):
            self.assertIn('out', system.custom_graphs)
        response = response['out']
        with self.subTest('Check consistency of outputs'):
            self.assertEqual(response, system.custom_graphs['out'])
        with self.subTest('Check that 10 indices were tracked'):
            self.assertEqual(10, len(response))
        recorded_indices = {summary.name for summary in response}
        with self.subTest('Check that 5 min keys were tracked'):
            min_keys = [3, 4, 8, 11, 14]
            for key in min_keys:
                self.assertIn(key, recorded_indices)
        with self.subTest('Check that 5 max keys were tracked'):
            max_keys = [7, 8, 24, 25, 27, 28]
            for key in max_keys:
                self.assertIn(key, recorded_indices)

    def test_specific_indices(self):
        instance_tracker = InstanceTracker(index='idx',
                                           metric='ce',
                                           n_max_to_keep=0,
                                           n_min_to_keep=0,
                                           list_to_keep=[1, 5, 17],
                                           outputs='out',
                                           mode='train')
        system = sample_system_object()
        instance_tracker.system = system
        response = Data()
        self._simulate_training(instance_tracker, response)

        # Check responses
        with self.subTest('Check that a response was written'):
            self.assertIn('out', response)
        with self.subTest('Check that data was written to the system'):
            self.assertIn('out', system.custom_graphs)
        response = response['out']
        with self.subTest('Check consistency of outputs'):
            self.assertEqual(response, system.custom_graphs['out'])
        with self.subTest('Check that 3 indices were tracked'):
            self.assertEqual(3, len(response))
        recorded_indices = {summary.name for summary in response}
        with self.subTest('Check that the correct keys were tracked'):
            target_keys = [1, 5, 17]
            for key in target_keys:
                self.assertIn(key, recorded_indices)

    def test_min_max_and_specific(self):
        instance_tracker = InstanceTracker(index='idx',
                                           metric='ce',
                                           list_to_keep=[1, 5, 17],
                                           outputs='out',
                                           mode='train')
        system = sample_system_object()
        instance_tracker.system = system
        response = Data()
        self._simulate_training(instance_tracker, response)

        # Check responses
        with self.subTest('Check that a response was written'):
            self.assertIn('out', response)
        with self.subTest('Check that data was written to the system'):
            self.assertIn('out', system.custom_graphs)
        response = response['out']
        with self.subTest('Check consistency of outputs'):
            self.assertEqual(response, system.custom_graphs['out'])
        with self.subTest('Check that 13 indices were tracked'):
            self.assertEqual(13, len(response))
        recorded_indices = {summary.name for summary in response}
        with self.subTest('Check that 5 min keys were tracked'):
            min_keys = [3, 4, 8, 11, 14]
            for key in min_keys:
                self.assertIn(key, recorded_indices)
        with self.subTest('Check that 5 max keys were tracked'):
            max_keys = [7, 8, 24, 25, 27, 28]
            for key in max_keys:
                self.assertIn(key, recorded_indices)
        with self.subTest('Check that the specified indices were tracked'):
            target_keys = [1, 5, 17]
            for key in target_keys:
                self.assertIn(key, recorded_indices)

    def test_save_and_load_state(self):
        def instantiate_system():
            tracker = InstanceTracker(index='idx',
                                      metric='ce',
                                      n_max_to_keep=0,
                                      n_min_to_keep=0,
                                      list_to_keep=[1, 5, 17],
                                      outputs='out',
                                      mode='train')
            system = sample_system_object()
            system.traces.append(tracker)
            tracker.system = system
            return system, tracker

        system, tracker = instantiate_system()

        # Make some changes
        response = Data()
        self._simulate_training(tracker, response)

        # Save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_dir=save_path)

        # reinstantiate system and load the state
        system, tracker = instantiate_system()
        system.load_state(save_path)

        loaded_tracker = system.traces[-1]
        response = loaded_tracker.index_history
        with self.subTest('Check that the restored container has the correct type'):
            self.assertTrue(isinstance(response, defaultdict))
        with self.subTest('Check that the mode was saved'):
            self.assertIn('train', response)
        response = response['train']
        with self.subTest('Check that all 3 labels have summaries'):
            self.assertEqual(3, len(response))
        with self.subTest('Check that the specified indices were restored'):
            target_keys = [1, 5, 17]
            for key in target_keys:
                self.assertIn(key, response)
        response = response[1]
        with self.subTest('Check that values were restored'):
            self.assertEqual(0.8, round(response[0][1], 5))
            self.assertEqual(0.3, round(response[1][1], 5))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import time
from collections import deque
from typing import Iterable, List, Set, Union

import numpy as np
from natsort import humansorted

from fastestimator.backend.get_lr import get_lr
from fastestimator.summary.summary import ValWithError
from fastestimator.summary.system import System
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import check_ds_id, check_io_names, parse_modes, to_list, to_number, to_set


@traceable()
class Trace:
    """Trace controls the training loop. Users can use the `Trace` base class to customize their own functionality.

    Traces are invoked by the fe.Estimator periodically as it runs. In addition to the current data dictionary, they are
    also given a pointer to the current `System` instance which allows access to more information as well as giving the
    ability to modify or even cancel training. The order of function invocations is as follows:

    ``` plot
            Training:                                       Testing:

        on_begin                                            on_begin
            |                                                   |
        on_epoch_begin (train)  <------<                    on_epoch_begin (test)  <------<
            |                          |                        |                         |
        on_batch_begin (train) <----<  |                    on_batch_begin (test) <----<  |
            |                       |  |                        |                      |  |
        on_batch_end (train) >-----^   |                    on_batch_end (test) >------^  |
            |                          ^                        |                         |
        on_epoch_end (train)           |                    on_epoch_end (test) >---------^
            |                          |                        |
        on_epoch_begin (eval)          |                    on_end
            |                          ^
        on_batch_begin (eval) <----<   |
            |                      |   |
        on_batch_end (eval) >-----^    |
            |                          |
        on_epoch_end (eval) >----------^
            |
        on_end
    ```

    Args:
        inputs: A set of keys that this trace intends to read from the state dictionary as inputs.
        outputs: A set of keys that this trace intends to write into the system buffer.
        mode: What mode(s) to execute this Trace in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Trace in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    system: System
    inputs: List[str]
    outputs: List[str]
    mode: Set[str]
    ds_id: Set[str]
    # You can put keys in here to have them automatically added to EvalEssential without the user having to manually add
    # them to the Estimator monitor_names. See BestModelSaver for an example.
    fe_monitor_names: Set[str]

    def __init__(self,
                 inputs: Union[None, str, Iterable[str]] = None,
                 outputs: Union[None, str, Iterable[str]] = None,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None) -> None:
        self.inputs = check_io_names(to_list(inputs))
        self.outputs = check_io_names(to_list(outputs))
        self.mode = parse_modes(to_set(mode))
        self.ds_id = check_ds_id(to_set(ds_id))
        self.fe_monitor_names = set()  # The use-case here is rare enough that we don't want to add this to the init sig

    def on_begin(self, data: Data) -> None:
        """Runs once at the beginning of training or testing.

        Args:
            data: A dictionary through which traces can communicate with each other or write values for logging.
        """
        pass

    def on_epoch_begin(self, data: Data) -> None:
        """Runs at the beginning of each epoch.

        Args:
            data: A dictionary through which traces can communicate with each other or write values for logging.
        """
        pass

    def on_batch_begin(self, data: Data) -> None:
        """Runs at the beginning of each batch.

        Args:
            data: A dictionary through which traces can communicate with each other or write values for logging.
        """
        pass

    def on_batch_end(self, data: Data) -> None:
        """Runs at the end of each batch.

        Args:
            data: The current batch and prediction data, as well as any information written by prior `Traces`.
        """
        pass

    def on_epoch_end(self, data: Data) -> None:
        """Runs at the end of each epoch.

        Args:
            data: A dictionary through which traces can communicate with each other or write values for logging.
        """
        pass

    def on_end(self, data: Data) -> None:
        """Runs once at the end training.

        Args:
            data: A dictionary through which traces can communicate with each other or write values for logging.
        """
        pass

    def get_outputs(self, ds_ids: Union[None, str, List[str]]) -> List[str]:
        """What outputs will be generated by this trace.

        You can ignore this unless you are designing a new trace that has special interactions between its outputs and
        particular dataset ids.

        Args:
            ds_ids: The ds_ids under which this trace will execute.

        Returns:
            The outputs of this trace.
        """
        return list(self.outputs)


@traceable()
class TrainEssential(Trace):
    """A trace to collect important information during training.

    Please don't add this trace into an estimator manually. FastEstimator will add it automatically.

    Args:
        monitor_names: Which keys from the data dictionary to monitor during training.
    """
    def __init__(self, monitor_names: Set[str]) -> None:
        super().__init__(inputs=monitor_names, mode="train", outputs=["steps/sec", "epoch_time", "total_time"])
        self.elapse_times = []
        self.train_start = None
        self.epoch_start = None
        self.step_start = None

    def on_begin(self, data: Data) -> None:
        self.train_start = time.perf_counter()
        data.write_with_log("num_device", self.system.num_devices)
        data.write_with_log("logging_interval", self.system.log_steps)

    def on_epoch_begin(self, data: Data) -> None:
        if self.system.log_steps:
            self.epoch_start = time.perf_counter()
            self.step_start = time.perf_counter()

    def on_batch_end(self, data: Data) -> None:
        if self.system.log_steps and (self.system.global_step % self.system.log_steps == 0
                                      or self.system.global_step == 1):
            for key in self.inputs:
                if key in data:
                    data.write_with_log(key, data[key])
            if self.system.global_step > 1:
                self.elapse_times.append(time.perf_counter() - self.step_start)
                data.write_with_log("steps/sec", round(self.system.log_steps / np.sum(self.elapse_times), 2))
            self.elapse_times = []
            self.step_start = time.perf_counter()

    def on_epoch_end(self, data: Data) -> None:
        if self.system.log_steps:
            self.elapse_times.append(time.perf_counter() - self.step_start)
            data.write_with_log("epoch_time", "{} sec".format(round(time.perf_counter() - self.epoch_start, 2)))

    def on_end(self, data: Data) -> None:
        self.system.mode = 'train'  # Set mode to 'train' for better log visualization
        data.write_with_log("total_time", "{} sec".format(round(time.perf_counter() - self.train_start, 2)))
        for model in self.system.network.models:
            if hasattr(model, "current_optimizer"):
                data.write_with_log(model.model_name + "_lr", get_lr(model))


@traceable()
class EvalEssential(Trace):
    """A trace to collect important information during evaluation.

    Please don't add this trace into an estimator manually. FastEstimator will add it automatically.

    Args:
        monitor_names: Any keys which should be collected over the course of an eval epoch.
    """
    def __init__(self, monitor_names: Set[str]) -> None:
        super().__init__(mode="eval", inputs=monitor_names)
        self.eval_results = None

    def on_epoch_begin(self, data: Data) -> None:
        self.eval_results = None

    def on_batch_end(self, data: Data) -> None:
        if self.eval_results is None:
            self.eval_results = {key: [data[key]] for key in self.inputs if key in data}
        else:
            for key in self.inputs:
                if key in data:
                    self.eval_results[key].append(data[key])

    def on_epoch_end(self, data: Data) -> None:
        for key, value_list in self.eval_results.items():
            data.write_with_log(key, np.mean(np.array(value_list), axis=0))


@traceable()
class TestEssential(Trace):
    """A trace to collect important information during evaluation.

    Please don't add this trace into an estimator manually. FastEstimator will add it automatically.

    Args:
        monitor_names: Any keys which should be collected over the course of an test epoch.
    """
    def __init__(self, monitor_names: Set[str]) -> None:
        super().__init__(mode="test", inputs=monitor_names)
        self.test_results = None

    def on_epoch_begin(self, data: Data) -> None:
        self.test_results = None

    def on_batch_end(self, data: Data) -> None:
        if self.test_results is None:
            self.test_results = {key: [data[key]] for key in self.inputs if key in data}
        else:
            for key in self.inputs:
                if key in data:
                    self.test_results[key].append(data[key])

    def on_epoch_end(self, data: Data) -> None:
        for key, value_list in self.test_results.items():
            data.write_with_log(key, np.mean(np.array(value_list), axis=0))


@traceable()
class Logger(Trace):
    """A Trace that prints log messages.

    Please don't add this trace into an estimator manually. FastEstimator will add it automatically.
    """
    def __init__(self) -> None:
        super().__init__(inputs="*")

    def on_begin(self, data: Data) -> None:
        if not self.system.mode == "test":
            start_step = 1 if not self.system.global_step else self.system.global_step
            self._print_message("FastEstimator-Start: step: {}; ".format(start_step), data)

    def on_batch_end(self, data: Data) -> None:
        if self.system.mode == "train" and self.system.log_steps and (self.system.global_step % self.system.log_steps
                                                                      == 0 or self.system.global_step == 1):
            self._print_message("FastEstimator-Train: step: {}; ".format(self.system.global_step), data)

    def on_epoch_end(self, data: Data) -> None:
        if self.system.mode == "train" and self.system.log_steps:
            self._print_message("FastEstimator-Train: step: {}; ".format(self.system.global_step), data, True)
        elif self.system.mode == "eval":
            self._print_message("FastEstimator-Eval: step: {}; ".format(self.system.global_step), data, True)
        elif self.system.mode == "test":
            self._print_message("FastEstimator-Test: step: {}; ".format(self.system.global_step), data, True)

    def on_end(self, data: Data) -> None:
        if not self.system.mode == "test":
            self._print_message("FastEstimator-Finish: step: {}; ".format(self.system.global_step), data)

    def _print_message(self, header: str, data: Data, log_epoch: bool = False) -> None:
        """Print a log message to the screen, and record the `data` into the `system` summary.

        Args:
            header: The prefix for the log message.
            data: A collection of data to be recorded.
            log_epoch: Whether epoch information should be included in the log message.
        """
        log_message = header
        if log_epoch:
            log_message += "epoch: {}; ".format(self.system.epoch_idx)
            self.system.write_summary('epoch', self.system.epoch_idx)
        deferred = []
        for key, val in humansorted(data.read_logs().items(), key=lambda x: x[0]):
            if isinstance(val, ValWithError):
                log_message += "{}: {}; ".format(key, str(val))
            else:
                val = to_number(val)
                if val.size > 1:
                    deferred.append("\n{}:\n{};".format(key, np.array2string(val, separator=',')))
                else:
                    log_message += "{}: {}; ".format(key, str(val))
            self.system.write_summary(key, val)
        log_message = log_message.strip()
        for elem in deferred:
            log_message += elem
        print(log_message)


def sort_traces(traces: List[Trace], ds_ids: List[str], available_outputs: Union[None, str,
                                                                                 Set[str]] = None) -> List[Trace]:
    """Sort traces to attempt to resolve any dependency issues.

    This is essentially a topological sort, but it doesn't seem worthwhile to convert the data into a graph
    representation in order to get the slightly better asymptotic runtime complexity.

    Args:
        traces: A list of traces (not inside schedulers) to be sorted.
        ds_ids: The ds_ids currently available to the traces.
        available_outputs: What output keys are already available for the traces to use. If None are provided, the
            sorting algorithm will assume that any keys not generated by traces are being provided by the system.
            This results in a less rigorous sorting.

    Returns:
        The sorted list of `traces`.

    Raises:
        AssertionError: If Traces have circular dependencies or require input keys which are not available.
    """
    sorted_traces = []
    trace_outputs = {output for trace in traces for output in trace.get_outputs(ds_ids=ds_ids)}
    if available_outputs is None:
        # Assume that anything not generated by a Trace is provided by the system
        available_outputs = {inp for trace in traces for inp in trace.inputs} - trace_outputs
        weak_sort = True
    else:
        available_outputs = to_set(available_outputs)
        weak_sort = False
    end_traces = deque()
    intermediate_traces = deque()
    intermediate_outputs = set()
    trace_deque = deque(traces)
    while trace_deque:
        trace = trace_deque.popleft()
        ins = set(trace.inputs)
        outs = set(trace.get_outputs(ds_ids=ds_ids))
        if not ins or isinstance(trace, (TrainEssential, EvalEssential, TestEssential)):
            sorted_traces.append(trace)
            available_outputs |= outs
        elif "*" in ins:
            if outs:
                end_traces.appendleft(trace)
            else:
                end_traces.append(trace)
        elif ins <= available_outputs or (weak_sort and (ins - outs - available_outputs).isdisjoint(trace_outputs)):
            sorted_traces.append(trace)
            available_outputs |= outs
        else:
            intermediate_traces.append(trace)
            intermediate_outputs |= outs

    already_seen = set()
    while intermediate_traces:
        trace = intermediate_traces.popleft()
        ins = set(trace.inputs)
        outs = set(trace.get_outputs(ds_ids=ds_ids))
        already_seen.add(trace)
        if ins <= available_outputs or (weak_sort and (ins - outs - available_outputs).isdisjoint(trace_outputs)):
            sorted_traces.append(trace)
            available_outputs |= outs
            already_seen.clear()
        elif ins <= (available_outputs | intermediate_outputs):
            intermediate_traces.append(trace)
        else:
            raise AssertionError("The {} trace has unsatisfiable inputs: {}".format(
                type(trace).__name__, ", ".join(ins - (available_outputs | intermediate_outputs))))

        if intermediate_traces and len(already_seen) == len(intermediate_traces):
            raise AssertionError("Dependency cycle detected amongst traces: {}".format(", ".join(
                [type(tr).__name__ for tr in already_seen])))
    sorted_traces.extend(list(end_traces))
    return sorted_traces


class PerDSTrace(Trace):
    def on_ds_begin(self, data: Data) -> None:
        """Runs at the beginning of each dataset.

        Args:
            data: A dictionary through which traces can communicate with each other. Output here will not be logged.
        """
        pass

    def on_ds_end(self, data: Data) -> None:
        """Runs at the beginning of each dataset.

        Args:
            data: A dictionary through which traces can communicate with each other. Output here will be accumulated
                across all available datasets and then logged during on_epoch_end.
        """
        pass
